# -*- coding: utf-8 -*-

"""
file: exporters.py

Description: some exporters for SEM documents

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import html
import json

import sem.storage
from sem.storage import Segmentation
from sem.storage import Span
from sem.storage import (
    tag_annotation_from_sentence as get_pos,
    chunk_annotation_from_sentence as get_chunks,
)
import sem.logger

try:
    from xml.etree import cElementTree as ET
except ImportError:
    from xml.etree import ElementTree as ET


def add_text(node, text):
    parts = text.split("\n")
    node.text = parts[0]
    for i in range(1, len(parts)):
        br = ET.SubElement(node, "lb")
        br.tail = "\n{}".format(parts[1])


def add_tail(node, tail):
    parts = tail.split("\n")
    node.tail = parts[0]
    for i in range(1, len(parts)):
        br = ET.SubElement(node, "lb")
        br.tail = "\n{}".format(parts[1])


class Exporter:
    def __init__(self, *args, **kwargs):
        pass

    def document_to_file(self, document, couples, output, encoding="utf-8", **kwargs):
        """
        write the document to a file in the given export format.

        Parameters
        ----------
            document : Document
                the document to export
            couples : dict (string -> string)
                the "entry name" <=> "entry index" that allows to
                retrieve information to export.
                ex: couples = {"chunking":"C", "NER":"N"}
            output : str
                the name of the file to write into
        """
        to_write = self.document_to_unicode(document, couples, **kwargs)
        try:
            output.write(to_write)
        except AttributeError:
            with open(output, "w", encoding=encoding, newline="") as output_stream:
                output_stream.write(to_write)

    def document_to_data(self, document, couples, **kwargs):
        """
        creates a new variable representing the document in the given
        export format.

        Parameters
        ----------
            document : Document
                the document to export
            couples : dict (string -> string)
                the "entry name" <=> "entry index" that allows to
                retrieve information to export.
                ex: couples = {"chunking":"C", "NER":"N"}
        """
        raise NotImplementedError(
            "export_to_data not implemented for class {}".format(self.__class__)
        )

    def document_to_unicode(self, document, couples, **kwargs):
        raise NotImplementedError(
            'document_to_unicode is not implemented for {}'.format(self.__class__)
        )


class BratExporter(Exporter):
    extension = "ann"

    def __init__(self, *args, **kwargs):
        pass

    def document_to_unicode(self, document, couples, **kwargs):
        lowers = dict([(x.lower(), y) for (x, y) in couples.items()])
        if "ner" not in lowers and "NER" in document.annotationsets:
            lowers["ner"] = "NER"
        if "ner" not in lowers:
            sem.logger.exception("No NER annotation specified for BRAT exporter")
        if not document.annotationset(lowers["ner"]):
            sem.logger.exception("No annotation %s in document", lowers["ner"])
        content = document.content
        parts = []
        for id, annotation in enumerate(
            document.annotationset(lowers["ner"]).get_reference_annotations(), 1
        ):
            parts.append(
                "T{id}\t{annotation.value} {annotation.lb} {annotation.ub}\t{txt}".format(
                    id=id,
                    annotation=annotation,
                    txt=content[annotation.lb: annotation.ub].replace("\r", "").replace("\n", " "),
                )
            )
        return "\n".join(parts)


class CoNLLExporter(Exporter):
    extension = "conll"

    def __init__(self, *args, **kwargs):
        self.scheme = kwargs.get("scheme", "BIO")

    def document_to_unicode(self, document, couples, **kwargs):
        if len(document.corpus.fields) == 0:
            sem.logger.warning("No fields found for Corpus, cannot create string.")
            return ""

        if (
            not couples
            or (len(couples) == 0)
            or (len(couples) == 1 and list(couples.keys())[0].lower() in ["word", "token"])
        ):
            return str(document.corpus)
        else:
            lower = {}
            fields = []
            for field in couples:
                lower[field.lower()] = couples[field]

            if "word" in lower:
                fields.append(lower["word"])
            elif "token" in lower:
                fields.append(lower["token"])
            else:
                all_keys = document.corpus.fields
                field = (
                    "word" if "word" in all_keys
                    else "token" if "token" in all_keys
                    else sorted(document.corpus.fields)[0]
                )
                fields.append(field)

            if "pos" in lower:
                fields.append(lower["pos"])
            if "chunking" in lower:
                fields.append(lower["chunking"])
            if "ner" in lower:
                fields.append(lower["ner"])

            for field in fields:
                if field not in document.corpus:
                    sem.logger.warning('field "%s" not in corpus, adding', field)
                    document.add_to_corpus(field, scheme=self.scheme)

            return document.corpus.unicode(fields)

    def document_to_data(self, document, couples, **kwargs):
        return document.corpus.sentences


class GateExporter(Exporter):
    extension = "gate.xml"

    def __init__(self, *args, **kwargs):
        pass

    def document_to_unicode(self, document, couples, **kwargs):
        return '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n' \
               + ET.tostring(
                   self.document_to_data(document, couples),
                   encoding="utf-8"
                ).decode("utf-8")

    def document_to_data(self, document, couples, **kwargs):
        gateDocument = ET.Element("GateDocument")
        gateDocument.set("version", "3")
        gateDocumentFeatures = ET.SubElement(gateDocument, "GateDocumentFeatures")
        # feature 1 : gate.SourceURL
        feature1 = ET.SubElement(gateDocumentFeatures, "Feature")
        name1 = ET.SubElement(feature1, "Name")
        name1.set("className", "java.lang.String")
        name1.text = "gate.SourceURL"
        value1 = ET.SubElement(feature1, "Value")
        value1.set("className", "java.lang.String")
        value1.text = "created from String"
        # feature 2 : MimeType
        feature2 = ET.SubElement(gateDocumentFeatures, "Feature")
        name2 = ET.SubElement(feature2, "Name")
        name2.set("className", "java.lang.String")
        name2.text = "MimeType"
        value2 = ET.SubElement(feature2, "Value")
        value2.set("className", "java.lang.String")
        value2.text = "text/plain"
        # feature 3 : docNewLineType
        feature3 = ET.SubElement(gateDocumentFeatures, "Feature")
        name3 = ET.SubElement(feature3, "Name")
        name3.set("className", "java.lang.String")
        name3.text = "docNewLineType"
        value3 = ET.SubElement(feature3, "Value")
        value3.set("className", "java.lang.String")
        value3.text = ""

        # The text with anchors
        textWithNodes = ET.SubElement(gateDocument, "TextWithNodes")
        content = document.content
        if "ner" not in couples and "NER" in document.annotationsets:
            couples["ner"] = "NER"
        if "ner" in couples:
            annotationset = document.annotationset(couples["ner"])
            annotations = annotationset.get_reference_annotations()
            boundaries = set()
            for annotation in annotations:
                boundaries.add(annotation.lb)
                boundaries.add(annotation.ub)
            boundaries = sorted(boundaries)
        else:
            annotationset = None
            annotations = []
            boundaries = []

        if boundaries != []:
            textWithNodes.text = content[: boundaries[0]]
            for nth, boundary in enumerate(boundaries[:-1]):
                node = ET.SubElement(textWithNodes, "Node")
                node.set("id", str(boundary))
                node.tail = content[boundary: boundaries[nth+1]]
            node = ET.SubElement(textWithNodes, "Node")
            node.set("id", str(boundaries[-1]))
            node.tail = content[boundaries[-1]: len(content)]
        else:
            textWithNodes.text = content

        if annotations != []:
            id = 1
            typeAnnotationSet = ET.SubElement(gateDocument, "AnnotationSet")
            typeAnnotationSet.set("Name", annotationset.name)
            for annot in annotations:
                annotation = ET.SubElement(typeAnnotationSet, "Annotation")
                annotation.set("Id", str(id))
                annotation.set("Type", annot.value)
                annotation.set("StartNode", str(annot.lb))
                annotation.set("EndNode", str(annot.ub))
                id += 1

        return gateDocument


class HTMLInlineExporter(Exporter):
    extension = "html"

    def __init__(self, lang="fr", lang_style="default.css", *args, **kwargs):
        self._lang = lang
        self._lang_style = lang_style

    def document_to_unicode(self, document, couples, encoding="utf-8", **kwargs):
        entry_names = {}
        for entry in couples:
            entry_names[entry.lower()] = couples[entry]

        key = entry_names.get("ner", None)
        if key is None or document.annotationset(key) is None:
            key = entry_names.get("chunking", None)
        if key is None or document.annotationset(key) is None:
            key = entry_names.get("pos", None)
        if key is None or document.annotationset(key) is None:
            raise KeyError("Cannot find any annotation for export.")

        content = document.original_content[:]

        position2html = {}
        annotations = document.annotationset(key).get_reference_annotations()
        for annotation in reversed(annotations):
            lb = annotation.lb
            ub = annotation.ub
            value = annotation.value

            if ub not in position2html:
                position2html[ub] = []
            position2html[ub].insert(0, "</span>")
            if lb not in position2html:
                position2html[lb] = []
            position2html[lb].append('<span id="{0}" title="{0}">'.format(value))

        for index in reversed(sorted(position2html.keys())):
            content = content[:index] + "".join(position2html[index]) + content[index:]

        content = content.replace(
            "<head>", '<head>\n<link rel="stylesheet" href="{0}" />'.format(self._lang)
        )
        content = content.replace(
            "<head>", '<head>\n<link rel="stylesheet" href="{0}" />'.format(self._lang_style)
        )

        return content


class HTMLExporter(Exporter):
    extension = "html"

    def __init__(self, lang="fr", lang_style="default.css", *args, **kwargs):
        self._lang = lang
        self._lang_style = lang_style

    def escape_tokens(self, corpus, token_entry):
        """
        Returns a list of (HTML-)escaped token given a corpus and
        an entry where to find the tokens.
        """
        escaped = []
        for sentence in corpus:
            escaped.append([])
            for element in sentence:
                escaped[-1].append(html.escape(element[token_entry]))
        return escaped

    def make_escaped_content(self, document):
        content = document.content
        tokens = document.segmentation("tokens")
        escaped_tokens = [html.escape(content[token.lb: token.ub]) for token in tokens]
        escaped_nontokens = [
            html.escape(content[tokens[i].ub: tokens[i + 1].lb]) for i in range(len(tokens) - 1)
        ]
        escaped_nontokens.insert(0, html.escape(content[0: tokens[0].lb]))
        escaped_nontokens.append(html.escape(content[tokens[-1].ub: len(content)]))

        return escaped_tokens, escaped_nontokens

    def document_to_unicode(self, document, couples, encoding="utf-8", **kwargs):
        entry_names = {}
        for entry in couples:
            entry_names[entry.lower()] = couples[entry]

        pos_html = []
        chunk_html = []
        ner_html = []

        current_key = entry_names.get("pos", "POS")
        if current_key and current_key in document.annotationsets:
            pos_html = self.add_annotation_document(document, current_key)
        current_key = entry_names.get("chunking", entry_names.get("chunk", "chunking"))
        if current_key and current_key in document.annotationsets:
            chunk_html = self.add_annotation_document(document, current_key)
        current_key = entry_names.get("ner", "NER")
        if current_key and current_key in document.annotationsets:
            ner_html = self.add_annotation_document(document, current_key)

        return self.makeHTML_document(document, pos_html, chunk_html, ner_html, encoding)

    def add_annotation_document(self, document, column):
        annotations = document.annotationset(column).get_reference_annotations()[::-1]
        content = document.content

        parts = []
        last = len(content)
        for annotation in annotations:
            parts.append(
                html.escape(content[annotation.ub: last])
                .replace("\n", "<br />\n")
                .replace("\r<br />", "<br />\r")
            )
            parts.append("</span>")
            parts.append(
                html.escape(content[annotation.lb: annotation.ub])
                .replace("\n", "<br />\n")
                .replace("\r<br />", "<br />\r")
            )
            parts.append('<span id="{0}" title="{0}">'.format(annotation.value))
            last = annotation.lb
        parts.append(
            html.escape(content[0:last]).replace("\n", "<br />\n").replace("\r<br />", "<br />\r")
        )

        new_content = "".join(parts[::-1])
        return new_content

    def makeHTML_document(self, document, pos, chunk, ner, output_encoding):
        def checked(number):
            # whether the tab is checked or not
            if 1 == number:
                return ' checked="true"'
            else:
                return ""

        css_tabs = "tabs.css"
        css_lang = self._lang_style

        # header + div that will contain tabs
        html_page = [
            """<html>
    <head>
        <meta charset="{0}" />
        <title>{1}</title>
        <link rel="stylesheet" href="{2}" />
        <link rel="stylesheet" href="{3}" />
    </head>
    <body>
        <div class="wrapper">
            <h1>{4}</h1>
            <div class="tab_container">""".format(
                output_encoding, document.name, css_tabs, css_lang, document.name
            )
        ]

        # the annotations that will be outputted
        nth = 1
        annots = []

        #
        # declaring tabs in HTML. TODO: refactor duped code
        #

        if pos != []:
            html_page.append(
                """
                <input id="tab{0}" type="radio" name="tabs"{1} />
                <label for="tab{0}">Part-Of-Speech</label>""".format(
                    nth, checked(nth)
                )
            )
            annots.append(
                """
                <section id="content{0}" class="tab-content">
{1}
                </section>""".format(
                    nth, pos
                )
            )
            nth += 1

        if chunk != []:
            html_page.append(
                """
                <input id="tab{0}" type="radio" name="tabs"{1} />
                <label for="tab{0}">Chunking</label>""".format(
                    nth, checked(nth)
                )
            )
            annots.append(
                """
                <section id="content{0}" class="tab-content">
{1}
                </section>""".format(
                    nth, chunk
                )
            )
            nth += 1

        if ner != []:
            html_page.append(
                """
                <input id="tab{0}" type="radio" name="tabs"{1} />
                <label for="tab{0}">Named Entity</label>""".format(
                    nth, checked(nth)
                )
            )
            annots.append(
                """
                <section id="content{0}" class="tab-content">
{1}
                </section>""".format(
                    nth, ner
                )
            )
            nth += 1

        # annotations are put after the tab declarations
        for annot in annots:
            html_page.append(annot)

        # closing everything that remains to be closed
        html_page.append(
            """
            </div>
        </div>
    </body>
</html>
"""
        )

        return "".join(html_page)

    def document_to_data(self, document, couples, **kwargs):
        """
        returns an ElementTree object (ATM) of the HTML page.
        """
        return ET.ElementTree(
            ET.fromstring(
                self.document_to_unicode(
                    document, couples, encoding=kwargs.pop("encoding", "utf-8"), **kwargs
                )
            )
        )


class JSONExporter(Exporter):
    extension = "json"

    def __init__(self, *args, **kwargs):
        pass

    def document_to_unicode(self, document, couples, **kwargs):
        return json.dumps(
            self.document_to_data(document, couples, **kwargs), indent=2, ensure_ascii=False
        )

    def document_to_data(self, document, couples, **kwargs):
        """
        This is just creating a dictionary from the document.
        Nearly copy-pasta of the Document.unicode method.
        """

        json_dict = {}
        json_dict["name"] = document.name
        json_dict["content"] = document.content
        json_dict["metadatas"] = document.metadatas

        json_dict["segmentations"] = {}
        for seg in document.segmentations.values():
            json_dict["segmentations"][seg.name] = {}
            ref = seg.reference.name if isinstance(seg.reference, Segmentation) else seg.reference
            if ref:
                json_dict["segmentations"][seg.name]["reference"] = ref
            json_dict["segmentations"][seg.name]["spans"] = [
                {"s": span.lb, "l": len(span)} for span in seg.spans
            ]

        json_dict["annotations"] = {}
        for annotation in document.annotationsets.values():
            json_dict["annotations"][annotation.name] = {}
            reference = (
                ""
                if not annotation.reference
                else (
                    annotation.reference
                    if isinstance(annotation.reference, str)
                    else annotation.reference.name
                )
            )
            if reference:
                json_dict["annotations"][annotation.name]["reference"] = reference
            json_dict["annotations"][annotation.name]["annotations"] = [
                {"v": tag.value, "s": tag.lb, "l": len(tag)} for tag in annotation
            ]

        return json_dict


class SEMExporter(Exporter):
    extension = "sem.xml"

    def __init__(self, *args, **kwargs):
        pass

    def document_to_file(self, document, couples, output, encoding="utf-8", **kwargs):
        with open(output, "w", encoding=encoding) as output_stream:
            # both SEM documents and SEM corpora have their write method,
            # this allows to have only one exporter for both
            document.write(output_stream, add_header=True)

    def document_to_data(self, document, couples, **kwargs):
        """
        This is just creating a dictionary from the document.
        Nearly copy-pasta of the Document.unicode method.
        """

        return document


class AnalecTEIExporter(Exporter):
    extension = "analec.tei.xml"

    def __init__(self, *args, **kwargs):
        pass

    def document_to_unicode(self, document, couples, **kwargs):
        return '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n' + ET.tostring(
            self.document_to_data(document, couples), encoding="utf-8"
        ).decode("utf-8")

    def document_to_data(self, document, couples, **kwargs):
        teiCorpus = ET.Element("teiCorpus")
        teiCorpus.set("xmlns", "http://www.tei-c.org/ns/1.0")
        teiHeader = ET.SubElement(teiCorpus, "teiHeader")
        fileDesc = ET.SubElement(teiHeader, "fileDesc")
        titleStmt = ET.SubElement(fileDesc, "titleStmt")
        title = ET.SubElement(titleStmt, "title")
        title.text = ""
        respStmt = ET.SubElement(titleStmt, "respStmt")
        resp = ET.SubElement(respStmt, "resp")
        resp.text = ""
        name = ET.SubElement(respStmt, "name")
        name.text = ""
        publicationStmt = ET.SubElement(fileDesc, "publicationStmt")
        publisher = ET.SubElement(publicationStmt, "publisher")
        publisher.text = ""
        sourceDesc = ET.SubElement(fileDesc, "sourceDesc")
        sourceDesc.text = ""
        TEI = ET.SubElement(teiCorpus, "TEI")
        teiHeader = ET.SubElement(TEI, "teiHeader")
        teiHeader.text = ""
        titleStmt = ET.SubElement(fileDesc, "titleStmt")
        title = ET.SubElement(titleStmt, "title")
        title.text = ""
        respStmt = ET.SubElement(titleStmt, "respStmt")
        resp = ET.SubElement(respStmt, "resp")
        resp.text = ""
        name = ET.SubElement(respStmt, "name")
        name.text = ""
        publicationStmt = ET.SubElement(fileDesc, "publicationStmt")
        publisher = ET.SubElement(publicationStmt, "publisher")
        publisher.text = ""
        sourceDesc = ET.SubElement(fileDesc, "sourceDesc")
        sourceDesc.text = ""

        root = ET.SubElement(TEI, "text")
        body = ET.SubElement(root, "body")

        lower = {}
        for field in couples:
            lower[field.lower()] = couples[field]
        annotations = set(document.annotationsets.keys())
        field = None
        if len(couples) == 1:
            field = lower[sorted(lower.keys())[0]]
        else:
            field = lower.get("ner", None) if lower.get("ner", None) in annotations else None
            if field is None:
                field = (
                    lower.get("chunking", None)
                    if lower.get("chunking", None) in annotations
                    else None
                )
        if field is None:
            raise ValueError("Could not determine the field to use for TEI export.")

        content = document.content
        paragraphs = (
            document.segmentation("paragraphs").get_reference_spans()
            if document.segmentation("paragraphs") is not None
            else [Span(0, len(content))]
        )
        NEs = document.annotationset(field).get_reference_annotations()
        values = set([entity.value for entity in NEs])

        nth = dict([(value, 0) for value in values])
        for paragraph in paragraphs:
            entities = [
                entity for entity in NEs if entity.lb >= paragraph.lb and entity.ub <= paragraph.ub
            ]
            p = ET.SubElement(body, "p")
            if len(entities) == 0:
                p.text = content[paragraph.lb: paragraph.ub]
            else:
                p.text = content[paragraph.lb: entities[0].lb]
                for i, entity in enumerate(entities):
                    nth[entity.value] += 1
                    entity_start = ET.SubElement(
                        p,
                        "anchor",
                        {
                            "xml:id": "u-{0}-{1}-start".format(entity.value, nth[entity.value]),
                            "type": "AnalecDelimiter",
                            "subtype": "UnitStart",
                        },
                    )
                    entity_start.tail = content[entity.lb: entity.ub]
                    entity_end = ET.SubElement(
                        p,
                        "anchor",
                        {
                            "xml:id": "u-{0}-{1}-end".format(entity.value, nth[entity.value]),
                            "type": "AnalecDelimiter",
                            "subtype": "UnitEnd",
                        },
                    )
                    if i < len(entities) - 1:
                        entity_end.tail = content[entity.ub: entities[i + 1].lb]
                    else:
                        entity_end.tail = content[entity.ub: paragraph.ub]

        back = ET.SubElement(root, "back")
        for value in sorted(values):
            spanGrp = ET.SubElement(back, "spanGrp")
            spanGrp.set("type", "AnalecUnit")
            spanGrp.set("n", value)
            i = 0
            for entity in [ent for ent in NEs if ent.value == value]:
                i += 1
                ET.SubElement(
                    spanGrp,
                    "span",
                    {
                        "xml:id": "u-{0}-{1}".format(value, i),
                        "from": "#u-{0}-{1}-start".format(value, i),
                        "to": "#u-{0}-{1}-end".format(value, i),
                        "ana": "#u-{0}-{1}-fs".format(value, i),
                    },
                )

        fvLib = ET.SubElement(back, "fvLib")
        fvLib.set("n", "AnalecElementProperties")
        nth = dict([(value, 0) for value in values])
        for i, entity in enumerate(NEs):
            nth[entity.value] += 1
            ET.SubElement(
                fvLib, "fs", {"xml:id": "u-{0}-{1}-fs".format(entity.value, nth[entity.value])}
            )

        return teiCorpus


class TEINPExporter(Exporter):
    extension = "analec.tei.xml"

    def __init__(self, *args, **kwargs):
        pass

    def document_to_unicode(self, document, couples, encoding="utf-8", **kwargs):
        return '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n' + ET.tostring(
            self.document_to_data(document, couples), encoding="utf-8"
        ).decode("utf-8")

    def document_to_data(self, document, couples, **kwargs):
        teiCorpus = ET.Element("teiCorpus")
        teiCorpus.set("xmlns", "http://www.tei-c.org/ns/1.0")
        teiHeader = ET.SubElement(teiCorpus, "teiHeader")
        fileDesc = ET.SubElement(teiHeader, "fileDesc")
        titleStmt = ET.SubElement(fileDesc, "titleStmt")
        title = ET.SubElement(titleStmt, "title")
        title.text = ""
        respStmt = ET.SubElement(titleStmt, "respStmt")
        resp = ET.SubElement(respStmt, "resp")
        resp.text = ""
        name = ET.SubElement(respStmt, "name")
        name.text = ""
        publicationStmt = ET.SubElement(fileDesc, "publicationStmt")
        publisher = ET.SubElement(publicationStmt, "publisher")
        publisher.text = ""
        sourceDesc = ET.SubElement(fileDesc, "sourceDesc")
        sourceDesc.text = ""
        TEI = ET.SubElement(teiCorpus, "TEI")
        teiHeader = ET.SubElement(TEI, "teiHeader")
        teiHeader.text = ""
        titleStmt = ET.SubElement(fileDesc, "titleStmt")
        title = ET.SubElement(titleStmt, "title")
        title.text = ""
        respStmt = ET.SubElement(titleStmt, "respStmt")
        resp = ET.SubElement(respStmt, "resp")
        resp.text = ""
        name = ET.SubElement(respStmt, "name")
        name.text = ""
        publicationStmt = ET.SubElement(fileDesc, "publicationStmt")
        publisher = ET.SubElement(publicationStmt, "publisher")
        publisher.text = ""
        sourceDesc = ET.SubElement(fileDesc, "sourceDesc")
        sourceDesc.text = ""

        root = ET.SubElement(TEI, "text")
        body = ET.SubElement(root, "body")

        lower = {}
        for field in couples:
            lower[field.lower()] = couples[field]
        chunking_field = None
        try:
            chunking_field = lower["chunking"]
        except KeyError:
            message = (
                'No "chunking" field was found, please check you have'
                " chunking information in your pipeline."
            )
            sem.logger.exception(message)
            raise KeyError(message)

        content = document.content
        pronoun2analec = {
            "CL": "PR_CL",
            "CLO": "PR_CL_O",
            "CLR": "PR_CL_R",
            "CLS": "PR_CL_S",
            "PRO": "PR_PRO",
            "PROREL": "PR_REL",
            "PROWH": "PR_WH",
            "P+PRO": "PR_PP",
        }
        words = document.segmentation("tokens").get_reference_spans()
        paragraphs = document.segmentation("paragraphs").get_reference_spans() or Span(
            0, len(content)
        )
        np_chunks = [
            annotation
            for annotation in document.annotationset(chunking_field)
            if annotation.value == "NP"
        ]
        pos_tags = document.annotationset(lower["pos"])[:]
        pos = []
        for i in range(len(np_chunks)):
            chunk = np_chunks[i]
            pos.append(
                [annot for annot in pos_tags if annot.lb >= chunk.lb and annot.ub <= chunk.ub]
            )

        for i in range(len(np_chunks)):
            np_chunks[i].ub = words[np_chunks[i].ub - 1].ub
            np_chunks[i].lb = words[np_chunks[i].lb].lb

        nth = 0
        for paragraph in paragraphs:
            nps = [
                chunk
                for chunk in np_chunks
                if chunk.lb >= paragraph.lb and chunk.ub <= paragraph.ub
            ]
            p = ET.SubElement(body, "p")
            if len(nps) == 0:
                p.text = content[paragraph.lb: paragraph.ub]
            else:
                p.text = content[paragraph.lb: nps[0].lb]
                for i, np in enumerate(nps):
                    nth += 1
                    np_start = ET.SubElement(
                        p,
                        "anchor",
                        {
                            "xml:id": "u-MENTION-{0}-start".format(nth),
                            "type": "AnalecDelimiter",
                            "subtype": "UnitStart",
                        },
                    )
                    np_start.tail = content[np.lb: np.ub]
                    np_end = ET.SubElement(
                        p,
                        "anchor",
                        {
                            "xml:id": "u-MENTION-{0}-end".format(nth),
                            "type": "AnalecDelimiter",
                            "subtype": "UnitEnd",
                        },
                    )
                    if i < len(nps) - 1:
                        np_end.tail = content[np.ub: nps[i + 1].lb]
                    else:
                        np_end.tail = content[np.ub: paragraph.ub]

        back = ET.SubElement(root, "back")
        spanGrp = ET.SubElement(back, "spanGrp")
        spanGrp.set("type", "AnalecUnit")
        spanGrp.set("n", "MENTION")
        for i, np in enumerate(np_chunks):
            ET.SubElement(
                spanGrp,
                "span",
                {
                    "xml:id": "u-MENTION-{0}".format(i + 1),
                    "from": "#u-MENTION-{0}-start".format(i + 1),
                    "to": "#u-MENTION-{0}-end".format(i + 1),
                    "ana": "#u-MENTION-{0}-fs".format(i + 1),
                },
            )

        fvLib = ET.SubElement(back, "fvLib")
        fvLib.set("n", "AnalecElementProperties")
        for i, np in enumerate(np_chunks):
            value = pronoun2analec.get(pos[i][0].value, "GN")

            fs = ET.SubElement(fvLib, "fs", {"xml:id": "u-MENTION-{0}-fs".format(i + 1)})
            f = ET.SubElement(fs, "f")
            f.set("name", "REF")
            ET.SubElement(f, "string")

            f = ET.SubElement(fs, "f")
            f.set("name", "CODE_SEM")
            fstring = ET.SubElement(f, "string")
            fstring.text = value

            f = ET.SubElement(fs, "f")
            f.set("name", "CATEGORIE")
            fstring = ET.SubElement(f, "string")
            fstring.text = value

        return teiCorpus


class REDENTEIExporter(Exporter):
    extension = "reden.tei.xml"

    def __init__(self, *args, **kwargs):
        pass

    def document_to_unicode(self, document, couples, **kwargs):
        return '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n' + ET.tostring(
            self.document_to_data(document, couples), encoding="utf-8"
        ).decode("utf-8")

    def document_to_data(self, document, couples, **kwargs):
        TEI = ET.Element("TEI")
        TEI.set("xmlns", "http://www.tei-c.org/ns/1.0")
        lang = document.metadata("lang")
        if lang is not None:
            TEI.set("xml:lang", lang)
        teiHeader = ET.SubElement(TEI, "teiHeader")
        fileDesc = ET.SubElement(teiHeader, "fileDesc")
        titleStmt = ET.SubElement(fileDesc, "titleStmt")
        title = ET.SubElement(titleStmt, "title")
        title.text = ""
        respStmt = ET.SubElement(titleStmt, "respStmt")
        resp = ET.SubElement(respStmt, "resp")
        resp.text = ""
        name = ET.SubElement(respStmt, "name")
        name.text = ""
        publicationStmt = ET.SubElement(fileDesc, "publicationStmt")
        publisher = ET.SubElement(publicationStmt, "publisher")
        publisher.text = ""
        sourceDesc = ET.SubElement(fileDesc, "sourceDesc")
        sourceDesc.text = ""
        titleStmt = ET.SubElement(fileDesc, "titleStmt")
        title = ET.SubElement(titleStmt, "title")
        title.text = ""
        respStmt = ET.SubElement(titleStmt, "respStmt")
        resp = ET.SubElement(respStmt, "resp")
        resp.text = ""
        name = ET.SubElement(respStmt, "name")
        name.text = ""
        publicationStmt = ET.SubElement(fileDesc, "publicationStmt")
        publisher = ET.SubElement(publicationStmt, "publisher")
        publisher.text = ""
        sourceDesc = ET.SubElement(fileDesc, "sourceDesc")
        sourceDesc.text = ""

        root = ET.SubElement(TEI, "text")
        body = ET.SubElement(root, "body")
        div = ET.SubElement(body, "div")

        lower = {}
        for field in couples:
            lower[field.lower()] = couples[field]
        annotations = set(document.annotationsets.keys())
        field = None
        if len(couples) == 1:
            field = lower[sorted(lower.keys())[0]]
        else:
            field = lower.get("ner", None) if lower.get("ner", None) in annotations else None
            if field is None:
                field = (
                    lower.get("chunking", None)
                    if lower.get("chunking", None) in annotations
                    else None
                )
        if field is None:
            raise ValueError("Could not determine the field to use for TEI export.")

        content = document.content
        paragraphs = (
            document.segmentation("paragraphs").get_reference_spans()
            if document.segmentation("paragraphs") is not None
            else [Span(0, len(content))]
        )
        NEs = document.annotationset(field).get_reference_annotations()

        for paragraph in paragraphs:
            entities = [
                entity for entity in NEs if entity.lb >= paragraph.lb and entity.ub <= paragraph.ub
            ]
            p = ET.SubElement(div, "p")
            if len(entities) == 0:
                p.text = content[paragraph.lb: paragraph.ub]
            else:
                p.text = content[paragraph.lb: entities[0].lb]
                for i, entity in enumerate(entities):
                    entity_xml = ET.SubElement(p, entity.value)
                    entity_xml.text = content[entity.lb: entity.ub]
                    if i < len(entities) - 1:
                        entity_xml.tail = content[entity.ub: entities[i + 1].lb]
                    else:
                        entity_xml.tail = content[entity.ub: paragraph.ub]

        return TEI


class TextExporter(Exporter):
    extension = "txt"

    def __init__(self, *args, **kwargs):
        pass

    def document_to_unicode(self, document, couples, **kwargs):
        corpus = document.corpus

        lower = {}
        for field in couples:
            lower[field.lower()] = couples[field]

        if "word" in lower:
            token_field = lower["word"]
        elif "token" in lower:
            token_field = lower["token"]
        elif "word" in corpus:
            token_field = "word"
        elif "token" in corpus:
            token_field = "token"
        else:
            raise RuntimeError("Cannot find token field")

        data = []
        for sentence in corpus:
            tokens = sentence.feature(token_field)

            if "pos" in lower and lower["pos"] in corpus:
                for annotation in get_pos(sentence, lower["pos"]):
                    tokens[annotation.ub - 1] += "/{}".format(annotation.value)
                    # regrouping tokens for tags spanning over >2 tokens
                    for i in range(annotation.lb, annotation.ub - 1):
                        tokens[i + 1] = "{}{}{}".format(tokens[i], "_", tokens[i + 1])
                        tokens[i] = ""

            if "chunking" in lower and lower["chunking"] in corpus:
                for annotation in get_chunks(sentence, lower["chunking"]):
                    tokens[annotation.lb] = "({0} {1}".format(
                        annotation.value, tokens[annotation.lb]
                    )
                    tokens[annotation.ub - 1] = "{0} )".format(tokens[annotation.ub - 1])

            if "ner" in lower and lower["ner"] in corpus:
                for annotation in get_chunks(sentence, lower["ner"]):
                    tokens[annotation.lb] = "({0} {1}".format(
                        annotation.value, tokens[annotation.lb]
                    )
                    tokens[annotation.ub - 1] = "{0} )".format(tokens[annotation.ub - 1])

            # if regrouping tokens, some are empty and would generate superfluous spaces
            tokens = [token for token in tokens if token != ""]
            data.append(" ".join(tokens[:]))
        return "\n".join(data)

    def document_to_data(self, document, couples, **kwargs):
        return self.document_to_unicode(document, couples).split("\n")


__exporters = {
    "brat": BratExporter,
    "conll": CoNLLExporter,
    "gate": GateExporter,
    "html": HTMLExporter,
    "html_inline": HTMLInlineExporter,
    "jason": JSONExporter,
    "sem_xml": SEMExporter,
    "tei_analec": AnalecTEIExporter,
    "tei_reden": REDENTEIExporter,
    "tei_np": TEINPExporter,
    "text": TextExporter
}


def get_exporter(name):
    return __exporters[name.lower()]


def available_exporters():
    return sorted(__exporters.keys())



# -*- coding:utf-8 -*-

"""file: map_annotations.py

Description: Map annotations according to a mapping. If no mapping is provided
for a given type, it will remain unchanged. If an empty mapping is provided,
every annotation of that type will be discarded. This module only affects
annotations, not CoNLL Corpus.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse

from sem.importers import read_conll
from sem.storage import compile_map


def main(argv=None):
    map_annotations(**vars(parser.parse_args(argv)))


def map_annotations(
    infile,
    mapping,
    outfile,
    column=-1,
    ienc=None,
    oenc=None,
    enc="utf-8",
    log_level="WARNING",
    log_file=None
):
    with open(mapping, encoding=ienc or enc) as input_stream:
        mapping = compile_map(input_stream)

    # here we do not simply go through the processor because of the non-fixed
    # columns that make using the processor a little bit awkward.
    # We instead emulate its behaviour here.
    with open(outfile, "w", encoding=oenc or enc) as output_stream:
        for sentence in read_conll(infile, encoding=ienc or enc):
            columns = sorted(sentence.keys())
            index = columns[column]
            tags = sentence.feature(index)
            prefixes = [
                ("_" if tag[0] == "_" else (tag[:2] if len(tag) > 2 and tag[1] == "-" else tag))
                for tag in tags
            ]
            values = [tags[i][len(prefixes[i]):] for i in range(len(tags))]
            values = [mapping.get(value, value) for value in values]
            values = [prefixes[i]+values[i] for i in range(len(tags))]
            sentence.add(values, index)
            output_stream.write(sentence.conll(columns) + "\n\n")


parser = argparse.ArgumentParser(
    "Map annotations according to a mapping."
    " If no mapping is provided for a given type,"
    " it will remain unchanged."
    " If an empty mapping is provided,"
    " every annotation of that type will be discarded.",
)

parser.add_argument("infile", help="The input file (CoNLL format)")
parser.add_argument("mapping", help="The mapping file")
parser.add_argument("outfile", help="The output file")
parser.add_argument(
    "-c", "--column", type=int, default=-1, help="The column to map (default: %(default)s)"
)
parser.add_argument("--input-encoding", dest="ienc", help="Encoding of the input (default: UTF-8)")
parser.add_argument("--output-encoding", dest="oenc", help="Encoding of the input (default: UTF-8)")
parser.add_argument(
    "--encoding",
    dest="enc",
    default="UTF-8",
    help="Encoding of both the input and the output (default: UTF-8)",
)
parser.add_argument(
    "-l",
    "--log",
    dest="log_level",
    choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
    default="WARNING",
    help="Increase log level (default: %(default)s)",
)
parser.add_argument("--log-file", dest="log_file", help="The name of the log file")



"""
file: annotation_gui.py

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import glob
import sys
import pathlib

import tkinter
import tkinter.filedialog
import tkinter.messagebox
import tkinter.scrolledtext
import tkinter.font
import tkinter.ttk

try:
    import pyperclip
except ImportError:
    pyperclip = None

import sem
import sem.util
from sem.constants import NUL
from sem.storage import SEMCorpus
from sem.storage import Tag, AnnotationSet
import sem.logger
import sem.importers
from sem.gui_components import (
    find_potential_separator,
    find_occurrences,
    from_tagset,
)
from sem.gui_components import (
    SEMTkTrainInterface,
    SearchFrame,
    SemTkLangSelector,
    SemTkResourceSelector,
)
import sem.pipelines


def update_annotations(document, annotation_name, annotations):
    annots = AnnotationSet(annotation_name)
    annots.annotations = annotations
    try:
        reference = document.annotationset(annotation_name).reference
    except Exception:
        reference = None
    document.add_annotationset(annots)
    if reference:
        document.set_reference(annotation_name, reference.name)


def check_in_tagset(tag, tagset):
    ok = tag in tagset
    if not ok:
        for the_tag in tagset:
            ok = the_tag.startswith(tag)
            if ok:
                break
    return ok


class AnnotationTool(tkinter.Frame):
    def __init__(self, parent, documents=None, tagset=None, *args, **kwargs):
        tkinter.Frame.__init__(self, parent, *args, **kwargs)

        parent.protocol("WM_DELETE_WINDOW", self.exit)

        self.resource_dir = sem.SEM_RESOURCE_DIR
        self.resource_to_load = None
        self.parent = parent
        self.user = None
        self._doc = None
        self.doc_is_modified = False
        self.annotation_name = None
        self.annotations = []
        self.annotations_tick = 0
        self._pipeline_loaded = False
        self._tagset_loaded = False

        # True = bind_all
        self.shortcuts = [
            ["Ctrl+o", ["open file", self.openfile_gui], [[self, True]]],
            ["Ctrl+Shift+o", ["open url", self.openurl], [[self, True]]],
            ["Ctrl+s", ["save", self.save], [[self, True]]],
            ["Ctrl+t", ["train", self.train], [[self, True]]],
            ["Ctrl+f", ["find", self.find_in_text], [[self, True]]],
            ["Ctrl+c", ["copy text", self.clipboard], [[self, True]]],
        ]

        self.lines_lengths = []

        #
        # menu
        #
        self.global_menu = tkinter.Menu(self.parent)
        # file menu
        self.file_menu = tkinter.Menu(self.global_menu, tearoff=False)
        self.global_menu.add_cascade(label="File", underline=0, menu=self.file_menu)
        self.file_menu.add_command(
            label="Open...", underline=0, command=self.openfile_gui, accelerator="Ctrl+O"
        )
        self.file_menu.add_command(
            label="Open url...", underline=5, command=self.openurl, accelerator="Ctrl+Shift+O"
        )
        self.file_menu.add_command(
            label="Save to...", underline=0, command=self.save, accelerator="Ctrl+S"
        )
        self.saveas_menu = tkinter.Menu(self.file_menu, tearoff=False)
        self.file_menu.add_cascade(label="Save as...", underline=5, menu=self.saveas_menu)
        self.saveas_menu.add_command(label="BRAT corpus", underline=0, command=self.save_brat)
        self.saveas_menu.add_command(label="GATE corpus", underline=0, command=self.save_gate)
        self.saveas_menu.add_command(
            label="TEI ANALEC corpus", underline=4, command=self.save_tei_analec
        )
        self.saveas_menu.add_command(
            label="TEI REDEN corpus", underline=4, command=self.save_tei_reden
        )
        self.saveas_menu.add_command(label="JSON corpus", underline=0, command=self.save_json)
        self.file_menu.entryconfig("Save to...", state=tkinter.DISABLED)
        self.file_menu.entryconfig("Save as...", state=tkinter.DISABLED)
        self.file_menu.add_separator()
        self.file_menu.add_command(
            label="Load tagset...", underline=5, command=self.load_tagset_gui
        )
        self.file_menu.add_command(
            label="Load workflow...", underline=5, command=self.load_workflow
        )
        self.file_menu.add_command(
            label="Load pipeline...", underline=5, command=self.load_pipeline
        )
        # edit menu
        self.edit_menu = tkinter.Menu(self.global_menu, tearoff=False)
        self.global_menu.add_cascade(label="Edit", underline=0, menu=self.edit_menu)
        self.edit_menu.add_command(label="Preferences...", command=self.preferences)
        # ? menu
        self.qm_menu = tkinter.Menu(self.global_menu, tearoff=False)
        self.global_menu.add_cascade(label="?", menu=self.qm_menu)
        self.qm_menu.add_command(label="About SEM...", command=self.about_sem)
        # final
        self.parent.config(menu=self.global_menu)

        self.new_type = tkinter.StringVar()

        self.bind_all("<Control-o>", self.openfile_gui)
        self.bind_all("<Control-O>", self.openurl)
        self.bind_all("<Control-s>", self.save)
        self.bind_all("<Control-t>", self.train)
        self.bind_all("<Control-f>", self.find_in_text)
        self.bind_all("<Control-c>", self.clipboard)
        self.focus_set()

        self.bind_all("<Tab>", self.tab)
        self.bind_all("<Shift-Tab>", self.shift_tab)

        self.available_chars_set = list("abcdefghijklmnopqrstuvwxyz") + [
            "F{}".format(i) for i in range(1, 13)
        ]
        self.SELECT_TYPE = "-- select type --"
        self.wish_to_add = []

        self.current_annotations = AnnotationSet("CurrentAnnotations")
        self.adder = None

        self.toolbar = tkinter.ttk.Frame(self)
        self.toolbar.pack(side="top", fill="x")

        self.train_btn = tkinter.ttk.Button(self.toolbar, text="train", command=self.train)
        self.train_btn.pack(side="left")
        self.train_btn.configure(state=tkinter.DISABLED)

        self.tag_document_btn = tkinter.ttk.Button(
            self.toolbar, text="tag document", command=self.tag_document
        )
        self.tag_document_btn.pack(side="left")
        self.tag_document_btn.configure(state=tkinter.DISABLED)

        self.type_combos = []
        self.add_type_lbls = []
        self.annotation_row = tkinter.ttk.PanedWindow(self, orient="horizontal")

        self.corpus_tree = tkinter.ttk.Treeview(self.annotation_row)
        self.corpus_tree_scrollbar = tkinter.ttk.Scrollbar(
            self.annotation_row, command=self.corpus_tree.yview
        )
        self.corpus_tree.configure(yscroll=self.corpus_tree_scrollbar.set)
        self.corpus_tree.heading("#0", text="corpus", anchor=tkinter.W)
        self.corpus_tree.bind("<<TreeviewSelect>>", self.load_document)
        self.corpus_documents = []
        self.corpus_id2doc = {}
        self.corpus_doc2id = {}

        self.text = tkinter.scrolledtext.ScrolledText(
            self.annotation_row, wrap=tkinter.WORD, font="Helvetica"
        )
        self.text.configure(state="disabled")
        self.text.bind("<Shift-Tab>", self.shift_tab)
        self.search = SearchFrame(self.text)
        for char in self.available_chars_set:
            self.text.bind("<{0}>".format(char), self.handle_char)
            self.parent.bind("<{0}>".format(char), self.handle_char)
            if char.islower():
                self.text.bind("<{0}>".format(char.upper()), self.handle_char)
                self.parent.bind("<{0}>".format(char.upper()), self.handle_char)

        self.tree = tkinter.ttk.Treeview(self.annotation_row)
        self.tree_scrollbar = tkinter.ttk.Scrollbar(self.annotation_row, command=self.tree.yview)
        self.tree.configure(yscroll=self.tree_scrollbar.set)
        self.tree.heading("#0", text="annotation sets", anchor=tkinter.W)
        self.tree.bind("<<TreeviewSelect>>", self.select_from_tree)
        self.tree_ids = {}
        self.annot2treeitems = {}

        self.annotation_row.add(self.corpus_tree)
        self.annotation_row.add(self.corpus_tree_scrollbar)
        self.annotation_row.add(self.text)
        self.annotation_row.add(self.tree)
        self.annotation_row.add(self.tree_scrollbar)
        self.annotation_row.pack(side="left", fill='both', expand=True)

        self.text.bind("<Button-1>", self.click)
        self.text.bind("<Delete>", self.delete)
        self.parent.bind("<Delete>", self.delete)
        self.text.bind("<Shift-Delete>", self.delete_all)
        self.parent.bind("<Shift-Delete>", self.delete_all)
        self.text.bind("<Escape>", self.unselect)
        self.parent.bind("<Escape>", self.unselect)

        self.tree.bind("<Delete>", self.delete)
        self.tree.bind("<Shift-Delete>", self.delete_all)

        # configuring a tag called SELECTION for selected named entities
        # bold_font = tkinter.font.Font(self.text)
        # bold_font.configure(weight="bold")
        # self.text.tag_configure("SELECTION", font=bold_font)
        ul_font = tkinter.font.Font(self.text)
        ul_font.configure(underline=1)
        self.text.tag_configure("SELECTION", font=ul_font)

        self.workflow = None

        self.position2annots = {}
        self.tree_ids = {}
        self.annot2treeitems = {}
        self.ner2history = {}

        # preferences
        self._whole_word = tkinter.BooleanVar()
        self._whole_word.set(True)
        self.wikinews_format = tkinter.BooleanVar()
        self.wikinews_format.set(False)

        self.pipeline = None

        if tagset:
            self.load_tagset(tagset)
        if documents:
            doc_list = []
            for doc in documents:
                doc_list.extend(glob.glob(doc))
            self.openfile(doc_list)
            ident = self.corpus_doc2id[self.corpus_documents[0].name]
            self.corpus_tree.selection_set(ident)
            self.corpus_tree.focus(ident)
            self.corpus_tree.see(ident)
            self.load_document()
        # skip_auth=> self.auth()

    @property
    def doc(self):
        return self._doc

    @doc.setter
    def doc(self, value):
        self._doc = value
        if self.can_label():
            self.tag_document_btn.configure(state=tkinter.NORMAL)
        else:
            self.tag_document_btn.configure(state=tkinter.DISABLED)

    @property
    def pipeline_loaded(self):
        return self._pipeline_loaded

    @pipeline_loaded.setter
    def pipeline_loaded(self, value):
        self._pipeline_loaded = value
        if self.can_label():
            self.tag_document_btn.configure(state=tkinter.NORMAL)
        else:
            self.tag_document_btn.configure(state=tkinter.DISABLED)

    @property
    def tagset_loaded(self):
        return self._tagset_loaded

    @tagset_loaded.setter
    def tagset_loaded(self, value):
        self._tagset_loaded = value
        if self.can_label():
            self.tag_document_btn.configure(state=tkinter.NORMAL)
        else:
            self.tag_document_btn.configure(state=tkinter.DISABLED)

    @property
    def whole_word(self):
        return bool(self._whole_word.get())

    def can_label(self):
        return self.pipeline_loaded and self.tagset_loaded and self.doc is not None

    def exit(self, event=None):
        do_quit = True

        if self.doc_is_modified:
            do_quit = tkinter.messagebox.askokcancel(
                "Quit",
                "The corpus has unsaved modifications, do you really want to quit?\n"
                "All changes will be lost."
            )

        if do_quit:
            self.destroy()
            self.parent.destroy()

    def auth(self, event=None):
        authTop = tkinter.Toplevel()
        authTop.grab_set()
        auth_login = tkinter.StringVar(authTop, value="admin")
        auth_pw = tkinter.StringVar(authTop, value="")

        def close():
            sys.exit(0)

        authTop.protocol("WM_DELETE_WINDOW", close)

        def check_auth():
            import time
            import hashlib

            pwd = auth_pw.get()
            if pwd == "":
                print("Please enter non empty password")
            try:
                content = open(".users").read()
                if (
                    "d033e22ae348aeb5660fc2140aec35850c4da997"
                    "aee5aca44055f2cd2f2ce4266909b69a5d96dad2\n" not in content
                ):
                    print(
                        "Something fishy about your .user file,"
                        " rewriting it with admin user only."
                    )
                    with open(".users", "w") as output_stream:
                        output_stream.write(
                            "d033e22ae348aeb5660fc2140aec35850c4da997"
                            "aee5aca44055f2cd2f2ce4266909b69a5d96dad2\n"
                        )
                    time.sleep(5.0)
                    sys.exit(1)
                h = hashlib.sha1()
                h.update(auth_login.get())
                login = h.hexdigest()
                h = hashlib.sha1()
                h.update(pwd)
                pw = h.hexdigest()
                checked = login + pw in content
                if checked:
                    self.user = auth_login.get()[:]
                    tkinter.messagebox.showinfo(
                        "Login success", "Logged succesfuly as {0}".format(self.user)
                    )

                    if self.user == "admin":
                        self.add_user_btn = tkinter.ttk.Button(
                            self.toolbar, text="add user", command=self.add_user
                        )
                        self.add_user_btn.pack(side="left")

                    authTop.destroy()
                else:
                    tkinter.messagebox.showerror("Login error", "Wrong login/password")
            except IOError:
                with open(".users", "w") as output_stream:
                    output_stream.write(
                        "d033e22ae348aeb5660fc2140aec35850c4da997"
                        "aee5aca44055f2cd2f2ce4266909b69a5d96dad2\n"
                    )
                print("Could not find .user file, rewriting it with admin user only.")

        authLabel = tkinter.Label(authTop, text="Enter credentials:")
        authLabel.grid(row=0, column=0, sticky=tkinter.W + tkinter.E, columnspan=2)

        tkinter.Label(authTop, text="login").grid(row=1, column=0, sticky=tkinter.W)
        auth_login_entry = tkinter.Entry(authTop, textvariable=auth_login)
        auth_login_entry.grid(row=1, column=1, sticky=tkinter.W)
        tkinter.Label(authTop, text="password").grid(row=2, column=0, sticky=tkinter.W)
        auth_pw_entry = tkinter.Entry(authTop, textvariable=auth_pw, show="*")
        auth_pw_entry.grid(row=2, column=1, sticky=tkinter.W)

        login_btn = tkinter.ttk.Button(authTop, text="login", command=check_auth)
        login_btn.grid(row=3, column=0, sticky=tkinter.W + tkinter.E, columnspan=2)

    def add_user(self, event=None):
        authTop = tkinter.Toplevel()
        authTop.grab_set()
        auth_login = tkinter.StringVar(authTop, value="admin")
        auth_pw = tkinter.StringVar(authTop, value="")

        def check_auth():
            import hashlib

            pwd = auth_pw.get()
            if pwd == "":
                print("Please enter non empty password")
                return
            try:
                lines = [line.strip() for line in open(".users").readlines()]
                if (
                    "d033e22ae348aeb5660fc2140aec35850c4da997"
                    "aee5aca44055f2cd2f2ce4266909b69a5d96dad2" not in lines
                ):
                    print(
                        "Something fishy about your .user file,"
                        " rewriting it with admin user only."
                    )
                    with open(".users", "w") as output_stream:
                        output_stream.write(
                            "d033e22ae348aeb5660fc2140aec35850c4da997"
                            "aee5aca44055f2cd2f2ce4266909b69a5d96dad2\n"
                        )
                h = hashlib.sha1()
                h.update(auth_login.get())
                login = h.hexdigest()
                h = hashlib.sha1()
                h.update(auth_pw.get())
                pw = h.hexdigest()
                for line in lines:
                    if line.startswith(login):
                        tkinter.messagebox.showerror(
                            "Cannot add user", "User {0} already exists".format(auth_login.get())
                        )
                        return

                with open(".users", "a") as output_stream:
                    output_stream.write("{0}{1}\n".format(login, pw))
                tkinter.messagebox.showinfo(
                    "New user", "Succesfuly added user {0}".format(auth_login.get())
                )
                authTop.destroy()
            except IOError:
                with open(".users", "w") as output_stream:
                    output_stream.write(
                        "d033e22ae348aeb5660fc2140aec35850c4da997"
                        "aee5aca44055f2cd2f2ce4266909b69a5d96dad2\n"
                    )
                print("Could not find .user file, rewriting it with admin user only.")

        authLabel = tkinter.Label(authTop, text="Enter credentials:")
        authLabel.grid(row=0, column=0, sticky=tkinter.W + tkinter.E, columnspan=2)

        tkinter.Label(authTop, text="login").grid(row=1, column=0, sticky=tkinter.W)
        auth_login_entry = tkinter.Entry(authTop, textvariable=auth_login)
        auth_login_entry.grid(row=1, column=1, sticky=tkinter.W)
        tkinter.Label(authTop, text="password").grid(row=2, column=0, sticky=tkinter.W)
        auth_pw_entry = tkinter.Entry(authTop, textvariable=auth_pw, show="*")
        auth_pw_entry.grid(row=2, column=1, sticky=tkinter.W)

        login_btn = tkinter.ttk.Button(authTop, text="login", command=check_auth)
        login_btn.grid(row=3, column=0, sticky=tkinter.W + tkinter.E, columnspan=2)

    #
    # file menu methods
    #

    def openfile(self, filenames):
        chunks_to_load = [self.annotation_name] if self.annotation_name else None

        documents = []
        names = set()
        for filename in filenames:
            if filename.endswith(".sem.xml") or filename.endswith(".sem"):
                try:
                    docs = sem.importers.sem_corpus_from_xml(
                        filename, chunks_to_load=chunks_to_load, load_subtypes=True
                    ).documents
                    for doc in docs:  # using reference annotations
                        for annotation_name in doc.annotationsets.keys():
                            doc.add_annotationset(
                                AnnotationSet(
                                    annotation_name,
                                    reference=None,
                                    annotations=doc.annotationset(
                                        annotation_name
                                    ).get_reference_annotations(),
                                )
                            )
                    for doc in [doc for doc in docs if doc.name not in names]:
                        documents.append(doc)
                        names.add(doc.name)
                except Exception:
                    doc = sem.importers.sem_document_from_xml(
                        filename, chunks_to_load=chunks_to_load, load_subtypes=True
                    )
                    if doc.name not in names:
                        names.add(doc.name)
                        documents.append(doc)
                        for annotation_name in documents[-1].annotationsets.keys():
                            documents[-1].add_annotationset(
                                AnnotationSet(
                                    annotation_name,
                                    reference=None,
                                    annotations=documents[-1]
                                    .annotationset(annotation_name)
                                    .get_reference_annotations(),
                                )
                            )
            else:
                doc = sem.importers.load(
                    filename, encoding="utf-8", tagset_name=self.annotation_name
                )
                if doc.name not in names:
                    documents.append(doc)
                    names.add(doc.name)

        if documents == []:
            return

        added = False
        for document in documents:
            tmp = self.add_document(document)
            added = added or tmp
        if not added:
            return

        self.load_document()

        self.train_btn.configure(state=tkinter.NORMAL)
        self.file_menu.entryconfig("Save to...", state=tkinter.NORMAL)
        self.file_menu.entryconfig("Save as...", state=tkinter.NORMAL)
        if self.adder is not None:
            self.adder.current_hierarchy_level = 0
            self.update_level()

    def openfile_gui(self, event=None):
        filenames = tkinter.filedialog.askopenfilenames(
            filetypes=[
                ("SEM readable files", (".txt", ".sem.xml", ".sem", ".ann", ".json")),
                ("text files", ".txt"),
                ("BRAT files", (".txt", ".ann")),
                ("SEM XML files", ("*.sem.xml", ".sem")),
                ("All files", ".*"),
            ]
        )
        if filenames == []:
            return
        self.openfile(filenames)

    def openurl(self, event=None):
        toplevel = tkinter.Toplevel()
        self.url = tkinter.StringVar()

        def cancel(event=None):
            self.url.set("")
            toplevel.destroy()

        def ok(event=None):
            document = sem.importers.from_url(
                self.url.get(), wikinews_format=bool(self.wikinews_format.get()), strip_html=True
            )
            if document is None:
                return

            added = self.add_document(document)
            if not added:
                return

            self.load_document()

            self.train_btn.configure(state=tkinter.NORMAL)
            self.file_menu.entryconfig("Save to...", state=tkinter.NORMAL)
            self.file_menu.entryconfig("Save as...", state=tkinter.NORMAL)
            cancel()

        label1 = tkinter.Label(toplevel, text="enter url:")
        label1.pack()

        text = tkinter.ttk.Entry(toplevel, textvariable=self.url)
        text.pack()
        text.focus_set()

        c = tkinter.ttk.Checkbutton(
            toplevel, text="Use Wikinews format", variable=self.wikinews_format
        )
        c.pack()

        toolbar = tkinter.ttk.Frame(toplevel)
        toolbar.pack(side="top", fill="x")
        ok_btn = tkinter.ttk.Button(toolbar, text="OK", command=ok)
        ok_btn.pack(side="left")
        cancel_btn = tkinter.ttk.Button(toolbar, text="cancel", command=cancel)
        cancel_btn.pack(side="left")
        toplevel.bind("<Return>", ok)
        toplevel.bind("<Escape>", cancel)

        toolbar.pack()

    def add_tag(self, value, start, end):
        if type(start) == int:
            start_pos = self.charindex2position(start)
        else:
            start_pos = start
        if type(end) == int:
            end_pos = self.charindex2position(end)
        else:
            end_pos = end

        pos = (start, end)
        self.position2annots[pos] = self.position2annots.get(pos, set())

        if value not in self.position2annots[pos]:
            self.text.tag_add(value, start_pos, end_pos)
            self.position2annots[pos].add(value)
            self.doc_is_modified = True

    def save(self, event=None):
        filename = tkinter.filedialog.asksaveasfilename(defaultextension=".sem.xml")
        if not filename:
            return

        self.unselect()
        if self.doc_is_modified:
            update_annotations(self.doc, self.annotation_name, self.current_annotations.annotations)

        corpus = SEMCorpus(documents=self.corpus_documents)
        with open(filename, "w", encoding="utf-8") as output_stream:
            corpus.write(output_stream)

        self.doc_is_modified = False

    def save_as_format(self, output_directory, fmt):
        if not output_directory:
            return
        output_path = pathlib.Path(output_directory)

        if self.doc_is_modified:
            update_annotations(self.doc, self.annotation_name, self.current_annotations.annotations)

        corpus = SEMCorpus(documents=self.corpus_documents)
        exporter = sem.exporters.get_exporter(fmt)()
        couples = {"ner": self.annotation_name}
        for document in corpus:
            name = document.name.replace(":", "")
            if not name.endswith(".txt"):
                name += ".txt"
            name = pathlib.Path(name)
            out_path = output_path / "{0}.{1}".format(name.stem, exporter.extension)
            if fmt == "brat":
                with open(output_path / name.name, "w", encoding="utf-8") as output_stream:
                    output_stream.write(document.content)
            exporter.document_to_file(document, couples, out_path, encoding="utf-8")

        self.doc_is_modified = False

    def save_brat(self, event=None):
        output_directory = tkinter.filedialog.askdirectory(initialdir=sem.SEM_DATA_DIR)
        self.save_as_format(output_directory, "brat")

    def save_gate(self, event=None):
        output_directory = tkinter.filedialog.askdirectory(initialdir=sem.SEM_DATA_DIR)
        self.save_as_format(output_directory, "gate")

    def save_tei_analec(self, event=None):
        output_directory = tkinter.filedialog.askdirectory(initialdir=sem.SEM_DATA_DIR)
        self.save_as_format(output_directory, "tei_analec")

    def save_tei_reden(self, event=None):
        output_directory = tkinter.filedialog.askdirectory(initialdir=sem.SEM_DATA_DIR)
        self.save_as_format(output_directory, "tei_reden")

    def save_json(self, event=None):
        output_directory = tkinter.filedialog.askdirectory(initialdir=sem.SEM_DATA_DIR)
        self.save_as_format(output_directory, "jason")

    #
    # Edit menu methods
    #

    def preferences(self, event=None):
        preferenceTop = tkinter.Toplevel()
        preferenceTop.focus_set()

        notebook = tkinter.ttk.Notebook(preferenceTop)

        frame1 = tkinter.ttk.Frame(notebook)
        notebook.add(frame1, text="general")
        frame2 = tkinter.ttk.Frame(notebook)
        notebook.add(frame2, text="shortcuts")

        c = tkinter.ttk.Checkbutton(
            frame1, text="Match whole word when broadcasting annotation", variable=self._whole_word
        )
        c.pack()

        shortcuts_vars = []
        cur_row = 0
        j = -1
        frame_list = []
        frame_list.append(tkinter.ttk.LabelFrame(frame2, text="common shortcuts"))
        frame_list[-1].pack(fill="both", expand="yes")
        for i, shortcut in enumerate(self.shortcuts):
            j += 1
            key, cmd, bindings = shortcut
            name, command = cmd
            shortcuts_vars.append(tkinter.StringVar(frame_list[-1], value=key))
            tkinter.Label(frame_list[-1], text=name).grid(row=cur_row, column=0, sticky=tkinter.W)
            entry = tkinter.Entry(frame_list[-1], textvariable=shortcuts_vars[j])
            entry.grid(row=cur_row, column=1)
            cur_row += 1
        notebook.pack()

    #
    # ? menu methods
    #

    def about_sem(self, event=None):
        options = {"background": "white"}
        aboutTop = tkinter.Toplevel(**options)
        aboutTop.title("About SEM")
        aboutTop.focus_set()

        two_cols = []
        two_cols.append(("author:", "Yoann Dupont"))
        two_cols.append(("mail:", "yoa.dupont@gmail.com"))
        two_cols.append(("website:", "http://www.lattice.cnrs.fr/sites/itellier/SEM.html"))
        two_cols.append(("github:", "https://github.com/YoannDupont/SEM"))
        two_cols.append(("online app:", "apps.lattice.cnrs.fr/sem"))

        x = 0
        label = tkinter.ttk.Label(aboutTop, text=sem.full_name(), **options)
        label.grid(row=x, column=0)
        x += 1
        tkinter.ttk.Label(aboutTop, text="", **options).grid(row=x, column=0)
        x += 1
        tkinter.ttk.Label(aboutTop, text="", **options).grid(row=x, column=0)
        x += 1
        for key, val in two_cols:
            label = tkinter.ttk.Label(aboutTop, text=key, **options)
            label.grid(row=x, column=0, sticky="w")
            label = tkinter.ttk.Label(aboutTop, text=val, **options)
            label.grid(row=x, column=1, sticky="w")
            x += 1
            tkinter.ttk.Label(aboutTop, text="", **options).grid(row=x, column=0)
            x += 1

    #
    # global methods
    #

    def train(self, event=None):
        if self.doc_is_modified:
            update_annotations(self.doc, self.annotation_name, self.current_annotations.annotations)

        SEMTkTrainInterface(self.corpus_documents)

    def handle_char(self, event):
        if self.adder is None:
            return
        # checks whether ctrl key is being pressed.
        # this prevents annotating when using a ctrl+key event.
        if event.state & 4:
            return
        if not self.adder.current_annotation:
            try:
                self.text.index("sel.first")
            except tkinter.TclError:
                return

        the_type = self.adder.type_from_letter(event.keysym)
        the_type = the_type or self.adder.type_from_letter(event.keysym.lower())
        if the_type is None:
            return
        if event.keysym.islower():
            fst = (
                self.charindex2position(self.adder.current_annotation.lb)
                if self.adder.current_annotation
                else self.text.index("sel.first")
            )
            lst = (
                self.charindex2position(self.adder.current_annotation.ub)
                if self.adder.current_annotation
                else self.text.index("sel.last")
            )
            self.wish_to_add = [self.adder.type_from_letter(event.keysym), fst, lst]
            self.add_annotation(event, remove_focus=False)
        else:
            if self.adder.current_annotation is not None:
                start = self.charindex2position(self.adder.current_annotation.lb)
                end = self.charindex2position(self.adder.current_annotation.ub)
            else:
                start, end = ("sel.first", "sel.last")
            try:
                for match in find_occurrences(
                    self.text.get(start, end), self.doc.content, whole_word=self.whole_word
                ):
                    cur_start = self.charindex2position(match.start())
                    cur_end = self.charindex2position(match.end())
                    if Tag(the_type, match.start(), match.end()) not in self.current_annotations:
                        self.wish_to_add = [the_type, cur_start, cur_end]
                        self.add_annotation(None, remove_focus=False)
            except tkinter.TclError:
                raise
            self.unselect()

    def add_annotation(self, event, remove_focus=True):
        cur_type = self.type_combos[self.adder.current_hierarchy_level].get()
        if cur_type.strip() != "" and cur_type != self.SELECT_TYPE:
            first = "sel.first"
            last = "sel.last"
            value = cur_type.split()[0]
        elif self.wish_to_add is not None:
            value, first, last = self.wish_to_add
        else:
            return

        try:
            if first == "sel.first":
                first = self.text.index("sel.first")
            if last == "sel.last":
                last = self.text.index("sel.last")
        except tkinter.TclError:  # no selection
            return

        self.doc_is_modified = True
        if self.adder.current_hierarchy_level == 0:
            pos = (self.position2charindex(first), self.position2charindex(last))
            greater = [
                annot
                for annot in self.current_annotations
                if annot.lb <= pos[0] and annot.ub >= pos[1] and value == annot.value
            ]
            tag = Tag(value, pos[0], pos[1])
            tag.levels = [value]
            if tag not in self.current_annotations and len(greater) == 0:
                self.text.tag_add(value, first, last)
                self.adder.current_annotation = tag
                index = 0
                for annot in self.current_annotations:
                    if annot.lb > tag.lb:
                        break
                    index += 1
                key = "{}".format(tag)
                item = self.tree.insert(
                    self.tree_ids[self.annotation_name],
                    index,
                    text='{0} "{1}" [{2}:{3}]'.format(
                        value, self.doc.content[pos[0]: pos[1]], pos[0], pos[1]
                    ),
                )
                self.treeitem2annot[item] = tag
                self.annot2treeitems[self.annotation_name][key] = item
                self.current_annotations.add(tag)
                item2 = self.tree.insert(
                    self.tree_ids["history"],
                    0,
                    text='{0} "{1}" [{2}:{3}]'.format(
                        value, self.doc.content[pos[0]: pos[1]], pos[0], pos[1]
                    ),
                )
                self.treeitem2annot[item2] = tag
                self.annot2treeitems["history"][key] = item2
                self.ner2history[item] = item2
                self.adder.current_annotation = tag

            self.text.tag_remove("SELECTION", "1.0", "end")
            self.type_combos[0].current(0)
        else:
            lb = self.position2charindex(first)
            ub = self.position2charindex(last)
            if self.wish_to_add is not None:
                for annot in self.current_annotations:
                    if annot.levels == []:
                        annot.levels = annot.value.split(".")
                    if (
                        annot.getLevel(0) == self.adder.current_annotation.getLevel(0)
                        and annot.lb == lb
                        and annot.ub == ub
                    ):
                        tree_item_str = self.locate_tree_item()
                        if check_in_tagset(annot.getValue(), self.tagset):
                            annot.setLevel(self.adder.current_hierarchy_level, self.wish_to_add[0])
                            new_text = '{0} "{1}" [{2}:{3}]'.format(
                                annot.getValue(), self.doc.content[annot.lb: annot.ub], lb, ub
                            )
                            self.tree.item(tree_item_str, text=new_text)
                            prev_item = self.ner2history.get(tree_item_str)
                            if prev_item:
                                self.tree.delete(prev_item)
                                del self.treeitem2annot[prev_item]
                            new_item = self.tree.insert(self.tree_ids["history"], 0, text=new_text)
                            self.ner2history[tree_item_str] = new_item
                            self.treeitem2annot[new_item] = self.treeitem2annot[tree_item_str]
                            key = "{}".format(Tag(annot.getLevel(0), lb, ub))
                            self.annot2treeitems["history"][key] = new_item
        if remove_focus:
            self.wish_to_add = None
            self.adder.current_annotation = None
            self.adder.current_hierarchy_level = 0
            self.update_level()
        else:
            self.text.tag_add("SELECTION", first, last)

    def click(self, event):
        if self.doc is None or self.adder is None:
            return

        self.text.tag_remove("SELECTION", "1.0", "end")
        prev_selection = self.adder.current_annotation
        self.adder.current_annotation = None
        self.wish_to_add = None
        index = event.widget.index("@{0},{1}".format(event.x, event.y))
        names = list(self.text.tag_names(index))
        charindex = self.position2charindex(index)
        try:
            names.remove("sel")
        except ValueError:
            pass

        annotations = [
            annotation
            for annotation in self.current_annotations
            if annotation.lb <= charindex and charindex <= annotation.ub
        ]

        if annotations != self.annotations:
            self.annotations = annotations
            self.annotations_tick = -1
        self.annotations_tick += 1
        if self.annotations_tick >= len(self.annotations):
            self.annotations_tick = 0
        if len(self.annotations) > 0:
            curr_annot = self.annotations[self.annotations_tick]
            ci2p = self.charindex2position
            self.text.tag_add("SELECTION", ci2p(curr_annot.lb), ci2p(curr_annot.ub))
            self.text.tag_remove(curr_annot.value, ci2p(curr_annot.lb), ci2p(curr_annot.ub))
            self.text.tag_add(curr_annot.value, ci2p(curr_annot.lb), ci2p(curr_annot.ub))
            self.adder.current_annotation = curr_annot

        tree_item = self.locate_tree_item()
        if tree_item is not None:
            self.tree.selection_set(tree_item)
            self.tree.focus(tree_item)
            self.select_from_tree()
            self.tree.see(tree_item)
        else:
            self.unselect()
        if len(annotations) == 0 or prev_selection != self.adder.current_annotation:
            self.adder.current_hierarchy_level = 0
            self.update_level()

    def locate_tree_item(self):
        if self.adder.current_annotation is None:
            return None
        if self.wish_to_add:
            lb = self.position2charindex(self.wish_to_add[1])
            ub = self.position2charindex(self.wish_to_add[2])
        else:
            lb = self.adder.current_annotation.lb
            ub = self.adder.current_annotation.ub
        ner_root = self.tree.get_children()[0]
        value = self.adder.current_annotation.getLevel(0)
        bounds = "[{}:{}]".format(lb, ub)
        for child in self.tree.get_children(ner_root):
            text = self.tree.item(child)["text"]
            ok = text.startswith(value) and text.endswith(bounds)
            if ok:
                return child
        return None

    def select_from_tree(self, event=None):
        parent = self.tree.parent(self.tree.selection()[0])
        if not parent:
            return

        annot = self.treeitem2annot[self.tree.selection()[0]]
        lb_str = self.charindex2position(annot.lb)
        ub_str = self.charindex2position(annot.ub)

        self.text.tag_remove("SELECTION", "1.0", "end")
        self.text.tag_add("SELECTION", lb_str, ub_str)
        self.adder.current_annotation = annot
        self.wish_to_add = None

        self.text.mark_set("insert", ub_str)
        self.text.see("insert")

    def unselect(self, event=None):
        self.text.tag_remove("SELECTION", "1.0", "end")
        self.wish_to_add = None
        if self.adder:
            self.adder.current_annotation = None
            self.adder.current_hierarchy_level = 0
            self.update_level()

    def delete(self, event):
        self.text.tag_remove("SELECTION", "1.0", "end")

        if self.adder.current_annotation is None:
            return

        value = self.adder.current_annotation.getValue()
        lb = self.adder.current_annotation.lb
        ub = self.adder.current_annotation.ub
        matching = [
            a
            for a in self.current_annotations
            if a.lb == lb and a.ub == ub and a.getValue() == value
        ]
        greater = [
            a
            for a in self.current_annotations
            if a.lb <= lb and a.ub >= ub and a.getValue() == value and a not in matching
        ]
        try:
            greater.remove(matching[0])
        except Exception:
            pass
        for annotation in matching:
            if len(greater) == 0:
                self.text.tag_remove(
                    self.adder.current_annotation.getLevel(0),
                    self.charindex2position(annotation.lb),
                    self.charindex2position(annotation.ub),
                )
            tag = Tag(self.adder.current_annotation.getLevel(0), annotation.lb, annotation.ub)
            key = "{}".format(tag)
            for v in self.annot2treeitems.values():
                item = v.get(key, None)
                if item is not None:
                    self.tree.delete(item)
                    if tag in self.current_annotations:
                        self.current_annotations.remove(self.adder.current_annotation)
                    del v[key]
                    del self.treeitem2annot[item]

        self.current_annotations.remove(self.adder.current_annotation)
        self.adder.current_annotation = None
        self.adder.current_hierarchy_level = 0
        self.update_level()
        self.doc_is_modified = True

    def delete_all(self, event):
        if self.adder.current_annotation is not None:
            value = self.adder.current_annotation.value
            start = self.adder.current_annotation.lb
            end = self.adder.current_annotation.ub
            for occ in find_occurrences(
                self.doc.content[start:end], self.doc.content, whole_word=False
            ):
                self.adder.current_annotation = Tag(value, occ.start(), occ.end())
                self.delete(event)
        self.adder.current_annotation = None
        self.adder.current_hierarchy_level = 0
        self.update_level()
        self.text.tag_remove("SELECTION", "1.0", "end")
        self.doc_is_modified = True

    def position2charindex(self, position):
        line, index = [int(e) for e in position.split(".")]
        return sum(self.lines_lengths[:line]) + index

    def charindex2position(self, charindex):
        lengths = self.lines_lengths
        cur = 0
        line = 1
        while cur + lengths[line] <= charindex:
            cur += lengths[line]
            line += 1
        offset = charindex - cur
        return "{0}.{1}".format(line, offset)

    def tab(self, event=None):
        if not self.tagset_loaded or self.doc is None:
            return
        self.adder.up_one_level()
        self.update_level()

    def shift_tab(self, event=None):
        if not self.tagset_loaded or self.doc is None:
            return
        self.adder.down_one_level()
        self.update_level()

    def update_level(self):
        for i in range(self.adder.max_depth()):
            if i != self.adder.current_hierarchy_level:
                self.type_combos[i].configure(state=tkinter.DISABLED)
            else:
                levels = (
                    self.adder.current_annotation.levels
                    if self.adder.current_annotation is not None
                    else []
                )
                subtrie = self.adder.shortcut_trie.goto(levels)
                keys = sorted([key for key in subtrie if key != NUL])
                for j in range(len(keys)):
                    shortcut = subtrie[keys[j]][NUL]
                    keys[j] += " ({0} or Shift+{0})".format(shortcut)
                self.type_combos[i]["values"] = [self.SELECT_TYPE] + keys
                self.type_combos[i].configure(state="readonly")

    def add_document(self, document):
        found = self.doc is not None and any(
            [document.name == doc.name for doc in self.corpus_documents]
        )
        if found:
            id = self.corpus_id2doc[self.doc.name]
            self.corpus_tree.selection_set(id)
            self.corpus_tree.focus(id)
            self.corpus_tree.see(id)
        else:
            id = self.corpus_tree.insert(
                "", len(self.corpus_tree.get_children()), text=document.name
            )
            self.corpus_id2doc[id] = document
            self.corpus_doc2id[document.name] = id
            self.corpus_documents.append(document)
            self.corpus_tree.selection_set(id)
            self.corpus_tree.focus(id)
            self.corpus_tree.see(id)
        return not found

    def load_document(self, event=None, same_doc=False):
        try:
            selection = self.corpus_tree.selection()[0]
        except IndexError:
            return

        document = self.corpus_id2doc[selection]
        if self.doc is None or (document.name != self.doc.name or same_doc):
            if self.doc is not None and self.doc_is_modified:
                update_annotations(
                    self.doc, self.annotation_name, self.current_annotations.annotations
                )

            self.doc = document

            previous_tree = self.tree_ids.get(self.annotation_name, None)
            if previous_tree:
                self.tree.delete(previous_tree)
            previous_tree = self.tree_ids.get("history", None)
            if previous_tree:
                self.tree.delete(previous_tree)

            if self.annotation_name is not None:
                self.tree_ids[self.annotation_name] = self.tree.insert(
                    "", len(self.tree_ids) + 1, text=self.annotation_name
                )
                self.tree_ids["history"] = self.tree.insert(
                    "", len(self.tree_ids) + 1, text="history"
                )
                self.annot2treeitems[self.annotation_name] = {}
            self.annot2treeitems["history"] = {}
            self.treeitem2annot = {}
            self.position2annots = {}
            self.current_annotations = AnnotationSet("CurrentAnnotations")
            if self.adder is not None:
                self.adder.current_annotation = None
            self.wish_to_add = None
            self.lines_lengths = [0] + [len(line) + 1 for line in self.doc.content.split("\n")]

            self.text.configure(state="normal")
            for tag_name in self.text.tag_names():
                self.text.tag_remove(tag_name, "1.0", "end")
            self.text.delete("1.0", "end")
            self.text.insert("end", self.doc.content)
            self.text.tag_remove("SELECTION", "1.0", "end")
            self.text.configure(state="disabled")

            if self.doc.annotationset(self.annotation_name):
                annots = self.doc.annotationset(self.annotation_name).get_reference_annotations()
                for nth_annot, annot in enumerate(annots):
                    annot.levels = annot.value.split(".")
                    self.add_tag(annot.levels[0], annot.lb, annot.ub)
                    item = self.tree.insert(
                        self.tree_ids[self.annotation_name],
                        len(self.annot2treeitems[self.annotation_name]) + 1,
                        text='{0} "{1}" [{2}:{3}]'.format(
                            annot.value, self.doc.content[annot.lb: annot.ub], annot.lb, annot.ub
                        ),
                    )
                    self.annot2treeitems[self.annotation_name]["{}".format(annot)] = item
                    self.treeitem2annot[item] = annot
                    self.current_annotations.add(annot)
                    if self.adder.shortcut_trie.goto(annot.levels[0]) is None:
                        separator = find_potential_separator(annot.value)
                        if separator is not None:
                            splitted = annot.value.split(separator)
                            for depth, type_to_add in enumerate(splitted, 0):
                                self.doc.annotationset(self.annotation_name)[nth_annot].setLevel(
                                    depth, type_to_add
                                )
                        else:
                            self.doc.annotationset(self.annotation_name)[nth_annot].setLevel(
                                0, annot.value
                            )
            self.doc_is_modified = False

    def load_tagset(self, filename):
        if self.doc and self.doc_is_modified and self.annotation_name:
            update_annotations(self.doc, self.annotation_name, self.current_annotations.annotations)

        tagset_name = pathlib.Path(filename).stem
        tagset = []
        with open(filename, "rU", encoding="utf-8") as input_stream:
            for line in input_stream:
                tagset.append(line.strip())
        tagset = [tag.split("#", 1)[0] for tag in tagset]
        tagset = [tag for tag in tagset if tag != ""]

        self.annotation_name = tagset_name
        self.tagset = set(tagset)
        self.tagset_loaded = True

        for combo in self.type_combos:
            combo.destroy()
        self.type_combos = []
        for add_type_lbl in self.add_type_lbls:
            add_type_lbl.destroy()
        self.add_type_lbls = [tkinter.ttk.Label(self.toolbar, text="add type:")]
        self.add_type_lbls[0].pack(side="left")

        for child in self.tree.get_children():
            self.tree.delete(child)
        self.tree_ids[self.annotation_name] = self.tree.insert(
            "", len(self.tree_ids) + 1, text=self.annotation_name
        )
        self.tree_ids["history"] = self.tree.insert("", len(self.tree_ids) + 1, text="history")
        self.annot2treeitems[self.annotation_name] = {}

        self.type_combos.append(tkinter.ttk.Combobox(self.toolbar))
        self.type_combos[0]["values"] = [self.SELECT_TYPE]
        self.type_combos[0].bind("<<ComboboxSelected>>", self.add_annotation)
        self.type_combos[0].pack(side="left")

        self.adder = from_tagset(tagset)
        for depth in range(self.adder.max_depth()):
            # label
            self.add_type_lbls.append(
                tkinter.ttk.Label(self.toolbar, text="add {0}type:".format("sub" * (depth)))
            )
            self.add_type_lbls[depth].pack(side="left")
            # combobox
            self.type_combos.append(tkinter.ttk.Combobox(self.toolbar))
            self.type_combos[depth]["values"] = [self.SELECT_TYPE]
            self.type_combos[depth].bind("<<ComboboxSelected>>", self.add_annotation)
            self.type_combos[depth].pack(side="left")
            for tag in sorted(set([t[depth] for t in self.adder.levels if len(t) > depth])):
                if len(self.type_combos) > 0:
                    self.type_combos[depth]["values"] = list(self.type_combos[depth]["values"]) + [
                        tag
                    ]
                if depth == 0:
                    self.text.tag_configure(tag, **self.adder.color[tag])
        self.update_level()
        self.doc = None
        self.load_document()

    def load_tagset_gui(self, event=None):
        filename = tkinter.filedialog.askopenfilename(
            filetypes=[("text files", ".txt"), ("All files", ".*")],
            initialdir=str(sem.SEM_DATA_DIR / "resources" / "tagsets"),
        )

        if len(filename) == 0:
            return

        self.load_tagset(filename)

    def find_in_text(self, event=None):
        self.search.find_in_text(event=event)

    def clipboard(self, event=None):
        if pyperclip is None:
            tkinter.messagebox.showwarning(
                "Copy to clipboard: missing library",
                "Cannot copy to clipboard: pyperclip library not installed."
            )
            return

        try:
            first = (
                self.charindex2position(self.adder.current_annotation.lb)
                if self.adder.current_annotation
                else self.text.index("sel.first")
            )
            last = (
                self.charindex2position(self.adder.current_annotation.ub)
                if self.adder.current_annotation
                else self.text.index("sel.last")
            )
            pyperclip.copy(self.text.get(first, last))
        except tkinter.TclError:  # triggers when no text is selected
            pass

    def load_resource(self, resource_kind, load):
        # dependency inversion, facilitates using workflow files and pipelines
        # in plans to remove workflow files later.
        top = tkinter.Toplevel()
        directory = sem.SEM_DATA_DIR / "resources" / resource_kind
        workflow_selector = SemTkResourceSelector(
            top, directory, filter=lambda x: True
        )
        lang_selector = SemTkLangSelector(top, directory)
        lang_selector.register(workflow_selector)
        vars_cur_row = 0
        vars_cur_row, _ = lang_selector.grid(row=vars_cur_row, column=0)
        vars_cur_row, _ = workflow_selector.grid(row=vars_cur_row, column=0)

        def cancel(event=None):
            top.destroy()

        def ok(event=None):
            load(workflow_selector.resource())
            top.destroy()

        ok_btn = tkinter.ttk.Button(top, text="load workflow", command=ok)
        ok_btn.grid(row=vars_cur_row, column=0)
        cancel_btn = tkinter.ttk.Button(top, text="cancel", command=cancel)
        cancel_btn.grid(row=vars_cur_row, column=1)

    def load_workflow(self, event=None):
        self.load_resource("workflow", self.load_workflowfile)

    def load_workflowfile(self, path):
        self.pipeline, _, _, _ = sem.pipelines.load_workflow(path)
        # self.tag_document_btn.configure(state=tkinter.NORMAL)
        self.pipeline_loaded = True

    def load_pipeline(self, event=None):
        self.load_resource("pipelines", self.load_pipelinefile)

    def load_pipelinefile(self, path):
        self.pipeline = sem.pipelines.load(path)
        # self.tag_document_btn.configure(state=tkinter.NORMAL)
        self.pipeline_loaded = True

    def tag_document(self, event=None):
        if self.pipeline is None or self.doc is None:
            return

        self.pipeline.process_document(self.doc)
        for key in self.doc.annotationsets:
            annotation = self.doc.annotationset(key)
            self.doc.add_annotationset(
                AnnotationSet(annotation.name, annotations=annotation.get_reference_annotations())
            )
        self.current_annotations = self.doc.annotationset(self.annotation_name)
        self.load_document(same_doc=True)
        self.doc_is_modified = True


def main(argv=None):
    annotation_gui(**vars(parser.parse_args(argv)))


def annotation_gui(documents=None, tagset=None, log_level="WARNING"):
    root = tkinter.Tk()
    root.title("SEM")
    sem.logger.setLevel(log_level)
    AnnotationTool(root, documents=documents, tagset=tagset).pack(expand=1, fill="both")
    root.mainloop()


parser = argparse.ArgumentParser("An annotation tool for SEM.")

parser.add_argument("-d", "--documents", nargs="*", help="Documents to load at startup.")
parser.add_argument("-t", "--tagset", help="The tagset to load at startup.")
parser.add_argument(
    "-l",
    "--log",
    dest="log_level",
    choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
    default="WARNING",
    help="Set the log level (default: %(default)s)",
)



# -*- coding: utf-8 -*-

"""file: storage.py

Description: defines multiple objects used for annotation. An annotation
is a set of values positioned using a segmentation.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import pathlib
import html
import re

import sem
import sem.util
import sem.logger
from sem.constants import BEGIN, IN, LAST, SINGLE, OUT, chunking_schemes
from sem.constants import NUL
from sem.CRF import Quark


class Span:
    """The Span object.

    Attributes
    ----------
    _lb : int
        the lower bound of a Span.
    _ub : int
        the upper bound of a Span.
    """

    __slots__ = ("_lb", "_ub")

    def __init__(self, lb, ub, length=-1):
        self._lb = min(lb, ub) if length < 0 else lb
        self._ub = max(lb, ub) if length < 0 else lb + length

    def __eq__(self, span):
        return self.lb == span.lb and self.ub == span.ub

    def __contains__(self, i):
        try:
            return self._lb <= i and i < self._ub
        except TypeError:
            return (self.lb <= i.lb) and (i.ub <= self.ub)

    def __len__(self):
        return self._ub - self._lb

    def __str__(self):
        return "[{span.lb}:{span.ub}]".format(span=self)

    @property
    def lb(self):
        return self._lb

    @property
    def ub(self):
        return self._ub

    @lb.setter
    def lb(self, lb):
        self._lb = min(lb, self._ub)

    @ub.setter
    def ub(self, ub):
        self._ub = max(ub, self._lb)

    def strictly_contains(self, i):
        return i > self._lb and i < self.ub

    def expand_lb(self, length):
        self._lb -= length

    def expand_ub(self, length):
        self._ub += length


def add_last(spanlist, span):
    """Appends "span" at the end of the list. If the last
    span's upper bound is equal to "span's" lower bound, the last
    span's upper bound is extended instead.
    """

    if span in spanlist[-1]:
        return

    if spanlist[-1].ub == span.lb:
        spanlist[-1].ub = span.ub
    else:
        spanlist.append(span)


class Tag:
    __slots__ = ("_span", "_value", "levels")

    def __init__(self, value, lb, ub, length=-1):
        self._span = Span(lb, ub, length=length)
        self._value = value.strip().strip(".")
        self.levels = self._value.split(".")

    def __len__(self):
        return len(self._span)

    def __eq__(self, tag):
        return tag is not None and self.value == tag.value and self.span == tag.span

    def __str__(self):
        return "{0},{1}".format(self.value, self.span)

    def __unicode__(self):
        return "{0},{1}".format(self.value, self.span)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value.strip().strip(".")
        self.levels = self.value.split(".")

    @property
    def span(self):
        return Span(self._span.lb, self._span.ub)

    @property
    def lb(self):
        return self._span.lb

    @lb.setter
    def lb(self, lb):
        self._span.lb = lb

    @property
    def ub(self):
        return self._span.ub

    @ub.setter
    def ub(self, ub):
        self._span.ub = ub

    def kind(self):
        return "chunking"

    def getLevel(self, nth):
        if nth >= len(self.levels):
            return ""
        return self.levels[nth]

    def setLevel(self, nth, value):
        while nth >= len(self.levels):
            self.levels.append("")
        self.levels[nth] = value
        while self.levels[-1] == "":
            del self.levels[-1]
        while len(self.levels) > nth + 1:
            del self.levels[-1]
        self._value = ".".join(self.levels).strip(".")

    def getValue(self):
        if self.levels == []:
            return self.value

        values = []
        do_it = True
        i = -1
        while do_it:
            i += 1
            value = self.getLevel(i)
            do_it = value != ""
            if do_it:
                values.append(value)
        return ".".join(values)


class AnnotationSet:
    def __init__(self, name, reference=None, annotations=None):
        self._name = name
        self._reference = reference
        if annotations is None:
            self._annotations = []
        else:
            self._annotations = annotations

    def __len__(self):
        return len(self._annotations)

    def __getitem__(self, i):
        return self._annotations[i]

    def __iter__(self):
        return iter(self._annotations)

    @property
    def name(self):
        return self._name

    @property
    def reference(self):
        return self._reference

    @property
    def annotations(self):
        return self._annotations

    @annotations.setter
    def annotations(self, annotations):
        self._annotations = annotations

    def add(self, annotation, after=None):
        i = 0
        if after is None:
            while i < len(self._annotations):
                if annotation.lb > self._annotations[i].lb:
                    None
                elif annotation.lb > self._annotations[i].ub:
                    None
                elif annotation.lb > self._annotations[i].ub:
                    break
                elif annotation.lb == self._annotations[i].lb:
                    if self._annotations[i].ub <= annotation.ub:
                        break
                else:
                    break
                i += 1
        else:
            while i < len(self._annotations):
                if self._annotations[i] == after:
                    i += 1
                    break
                i += 1
        self._annotations.insert(i, annotation)

    def append(self, annotation):
        self._annotations.append(annotation)

    def extend(self, annotations):
        self._annotations.extend(annotations)

    def remove(self, annotation):
        try:
            self._annotations.remove(annotation)
        except ValueError:  # annotation not in annotations
            pass

    def sort(self):
        self._annotations.sort(key=lambda x: (x.lb, -x.ub, x.value))

    def get_reference_annotations(self):
        if self.reference is None:
            return [Tag(a.value, a.lb, a.ub) for a in self.annotations]
        else:
            reference_spans = self.reference.get_reference_spans()
            return [
                Tag(
                    element.value,
                    reference_spans[element.lb].lb,
                    reference_spans[element.ub - 1].ub,
                )
                for element in self.annotations
            ]


def get_top_level(annotations):
    result = annotations[:]
    modified = True
    while modified:
        modified = False
        for i in range(len(result) - 1):
            modified = result[i].lb <= result[i + 1].lb and result[i].ub > result[i + 1].lb
            if modified:
                del result[i + 1]
                break
    return result


def get_bottom_level(annotations):
    result = annotations[:]
    modified = True
    while modified:
        modified = False
        for i in range(len(result) - 1):
            modified = result[i].lb <= result[i + 1].lb <= result[i].ub
            if modified:
                del result[i]
                break
    return result


str2filter = {"top level": get_top_level, "bottom level": get_bottom_level}


def chunks_to_annotation(lst, shift=0, strict=False):
    annotation = AnnotationSet("")
    start = 0
    length = 0
    value = ""
    last = len(lst) - 1
    for index, tag in enumerate(lst):
        flag = tag[0]
        if flag in OUT:
            if value != "":  # we just got out of a chunk
                annotation.append(Tag(value, start + shift, 0, length=length))
            value = ""
            length = 0
        elif flag in BEGIN:
            if value != "":  # begin after non-empty chunk ==> add annnotation
                annotation.append(Tag(value, start + shift, 0, length=length))
            value = tag[2:]
            start = index
            length = 1
            if index == last:  # last token ==> add annotation
                annotation.append(Tag(value, start + shift, 0, length=length))
        elif flag in IN:
            if value != tag[2:] and strict:
                raise ValueError(
                    'Got different values for same chunk: "{}" <> "{}"'.format(tag[2:], value)
                )
            length += 1
            if index == last:  # last token ==> add annotation
                annotation.append(Tag(value, start + shift, 0, length=length))
        elif flag in LAST:
            annotation.append(Tag(value, start + shift, 0, length=length + 1))
            value = ""
            length = 0
        elif flag in SINGLE:
            if value != "":  # begin after non-empty chunk ==> add annnotation
                annotation.append(Tag(value, start + shift, 0, length=length))
                value = ""
                length = 0
            annotation.append(Tag(tag[2:], index + shift, 0, length=1))
    return annotation


def chunk_annotation_from_sentence(sentence, column, shift=0, strict=False):
    return chunks_to_annotation(sentence.feature(column), shift=shift, strict=strict)


def chunk_annotation_from_corpus(corpus, column, name, reference=None, strict=False):
    """Return an annotation from a sentence. The annotation has to have one
    of the following tagging schemes:
       - BIO (Begin In Out)
       - BILOU (Begin In Last Out Unit-length)
       - BIOES (Begin In Out End Single)

    we define a general approach to handle the three at the same time.
    """

    annotation = AnnotationSet(name, reference=reference)
    shift = 0
    for sentence in corpus:
        annotation.extend(
            chunk_annotation_from_sentence(sentence, column, shift=shift, strict=strict).annotations
        )
        shift += len(sentence)
    return annotation


def tag_annotation_from_sentence(sentence, column, shift=0, strict=False):
    def is_begin(tag):
        return tag[0] != "_" or tag.startswith("__")

    annotation = AnnotationSet("")
    start = 0
    length = 0
    value = ""
    last = len(sentence) - 1
    for index, tag in enumerate(sentence.feature(column)):
        if is_begin(tag):
            if value != "":  # begin after non-empty chunk ==> add annnotation
                annotation.append(Tag(value, start + shift, 0, length=length))
            value = tag
            start = index
            length = 1
            if index == last:  # last token ==> add annotation
                annotation.append(Tag(value, start + shift, 0, length=length))
        else:
            if value != tag[1:]:
                if strict:
                    raise ValueError(
                        'Got different values for same POS: "{}" <> "{}"'.format(tag[1:], value)
                    )
                else:
                    value = tag[1:]  # most probable tag at the end.
            length += 1
            if index == last:  # last token ==> add annotation
                annotation.append(Tag(value, start + shift, 0, length=length))
    return annotation


def tag_annotation_from_corpus(corpus, column, name, reference=None, strict=False):
    """Return an annotation from a sentence. The annotation has the following
    scheme:
        add "_" at the beginning of an annotation if it "continues"
        the previous tag. It is the same as BIO, "B-" is replaced by None
        and "I-" by "_".
    """
    annotation = AnnotationSet(name, reference=reference)
    shift = 0
    for sentence in corpus:
        annotation.extend(
            tag_annotation_from_sentence(sentence, column, shift=shift, strict=strict).annotations
        )
        shift += len(sentence)
    return annotation


def annotation_from_sentence(sentence, column, shift=0, strict=False):
    """Return an AnnotationSet object for sentence. Checks sentence before
    calling either tag_annotation_from_sentence
    or chunk_annotation_from_sentence
    """
    flags = BEGIN + IN + LAST + SINGLE + OUT
    if all(token[0] in flags for token in sentence.feature(column)):
        return chunk_annotation_from_sentence(sentence, column, shift=shift, strict=strict)
    else:
        return tag_annotation_from_sentence(sentence, column, shift=shift, strict=strict)


_train_set = set(["train", "eval", "evaluate", "evaluation"])
_train = "train"
_label_set = set(["label", "annotate", "annotation"])
_label = "label"
_modes = _train_set | _label_set
_equivalence = dict(
    [[mode, _train] for mode in _train_set] + [[mode, _label] for mode in _label_set]
)
_allowed = dict(
    [[mode, set([_train, _label])] for mode in _train_set]
    + [[mode, set([_label])] for mode in _label_set]
)


class Entry:
    """The Entry object. It represents a field's identifier in a CoNLL corpus.
    An Entry may be used only in certain circumstances: for example, the
    output tag may only appear in train mode.
    """

    def __init__(self, name, mode="label"):
        if mode not in _modes:
            raise ValueError("Unallowed mode for entry: {0}".format(mode))
        self._name = name
        self._mode = _equivalence[mode]

    def __eq__(self, other):
        return self.name == other.name

    @property
    def name(self):
        return self._name

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        self._mode = _equivalence[mode]

    @property
    def is_train(self):
        return self.mode == _train

    @property
    def is_label(self):
        return self.mode == _label

    @staticmethod
    def fromXML(xml_element):
        return Entry(**xml_element.attrib)

    def has_mode(self, mode):
        return self.mode in _allowed[mode]


class Sentence:
    __slots__ = ("_features")

    def __init__(self, features=None):
        self._features = features or {}
        if self._features:
            iterator = iter(self._features.items())
            lst = len(next(iterator)[1])
            for key, value in self._features.items():
                if len(value) != lst:
                    raise ValueError("Invalid sentence: features of different lengths.")

    def __len__(self):
        keys = self.keys()
        if keys:
            return len(self._features[list(keys)[0]])
        return 0

    def conll(self, keys):
        return "\n".join(
            "\t".join(str(token) for token in seq)
            for seq in zip(*[self.feature(key) for key in keys])
        )

    def keys(self):
        return self._features.keys()

    def add(self, feature, key):
        """Add a feature to the sentence by mapping the feature function
        to the sentence.

        Parameters
        ----------
        feature : function Sentence -> list or list
            the feature function to apply to the Sentence or the list of values
        key : str
            The key where to put the output of the feature function
        """
        try:
            feature.__call__
        except AttributeError:  # not callable
            value = feature
        else:
            value = feature(self)
        self._features[key] = value

    def update(self, features_keys):
        """Update Sentence with every feature given in argument.

        Parameters
        ----------
        features_keys : List[(function Sentence -> list, str)]
            the list of "feature function" and "key" pairs to add the Sentence.
        """
        for feature, key in features_keys:
            self.add(feature, key)

    def feature(self, name):
        return self._features[name]

    def remove(self, key):
        del self._features[key]


class Corpus:
    def __init__(self, sentences=None, fields=None):
        self.sentences = sentences or []
        self._fields = fields

    def __contains__(self, item):
        return item in self.fields

    def __len__(self):
        return len(self.sentences)

    def __iter__(self):
        for element in self.sentences:
            yield element

    def __unicode__(self):
        return self.unicode(self.fields)

    @property
    def fields(self):
        return self._fields or self.sentences[0].keys()

    @fields.setter
    def fields(self, value):
        missing = sorted(set(value) - set(self.sentences[0].keys()))
        if missing:
            raise KeyError("missing fields: {}".format(','.join(missing)))
        self._fields = value

    def unicode(self, fields, separator="\t"):
        sentences = []
        for sentence in self:
            sentences.append([])
            for token in zip(*[sentence.feature(key) for key in fields]):
                sentences[-1].append(("\t".join(token)) + "\n")
        return "\n".join(["".join(sentence) for sentence in sentences])

    def is_empty(self):
        return 0 == len(self.sentences)

    def has_key(self, key):
        return key in self.fields

    def from_sentences(self, sentences, field_name="word"):
        del self.sentences[:]

        for sentence in sentences:
            self.sentences.append({field_name: sentence})

    def from_segmentation(self, content, tokens, sentences, field_name="word"):
        self.sentences = [
            Sentence({
                field_name: [
                    content[token.lb: token.ub]
                    for token in tokens.spans[sentence.lb: sentence.ub]
                ]
            })
            for sentence in sentences
        ]
        self.fields = [field_name]

    def write(self, fd, fields=None):
        for sentence in self:
            for token in zip(*[sentence.feature(key) for key in (fields or self.fields)]):
                fd.write(("\t".join(str(item) for item in token)) + "\n")
            fd.write("\n")


def compile_token(iterator):
    tokens = set()
    for item in iterator:
        item = item.split("#", 1)[0].strip()
        if item:
            tokens.add(item)
    return tokens


def compile_multiword(iterator):
    trie = Trie()
    for item in iterator:
        item = item.split("#", 1)[0].strip()
        if item:
            seq = item.split()
            trie.add(seq)
    return trie


def compile_map(iterator):
    out_map = {}
    for item in iterator:
        item = item.strip()
        if item != "":
            try:
                key, value = item.split("\t")
            except ValueError:
                key = item
                value = ""
            out_map[key] = value
    return out_map


class Segmentation:
    """Segmentation is just a holder for bounds. Those bounds can be word
    bounds or sentence bounds for example.
    By itself, it is not very useful, it become good in the context of
    a document for which it hold minimum useful information
    """

    def __init__(self, name, reference=None, spans=None):
        """parameters
        ----------
        name: unicode
            the name of the segmentation (tokens, sentences, paragraphs, etc.)
        reference: unicode or Segmentation
            if unicode: the name of the referenced segmentation in the document
            if Segmentation: the referenced segmentation
        spans: list of span
        bounds: list of span
        """

        self._name = name
        self._document = None
        self._reference = reference
        self._spans = spans

    def __len__(self):
        return len(self.spans)

    def __getitem__(self, i):
        return self._spans[i]

    def __iter__(self):
        for element in self.spans:
            yield element

    def append(self, span):
        if self._spans is None:
            self._spans = []
        self._spans.append(span)

    @property
    def name(self):
        return self._name

    @property
    def reference(self):
        return self._reference

    @property
    def spans(self):
        return self._spans

    def get_reference_spans(self):
        """returns spans according to the reference chain."""

        if self.reference is None:
            return self.spans
        else:
            reference_spans = self.reference.get_reference_spans()
            return [
                Span(reference_spans[element.lb].lb, reference_spans[element.ub - 1].ub)
                for element in self.spans
            ]


class Document:
    def __init__(
        self, name, content=None, encoding=None, lang=None, mime_type=None, corpus=None,
        original_content=None, **kwargs
    ):
        self._name = name
        self._content = content
        self.original_content = original_content
        self._segmentations = {}
        self._annotationsets = {}
        self._corpus = corpus or Corpus()
        self._metadatas = {}
        if encoding is not None:
            self._metadatas["encoding"] = encoding
        if lang is not None:
            self._metadatas["lang"] = lang
        if mime_type is not None:
            self._metadatas["MIME"] = mime_type

    @property
    def name(self):
        return self._name

    @property
    def content(self):
        return self._content

    @property
    def corpus(self):
        return self._corpus

    @content.setter
    def content(self, content):
        self._content = content

    @property
    def segmentations(self):
        return self._segmentations

    @property
    def annotationsets(self):
        return self._annotationsets

    @property
    def metadatas(self):
        return self._metadatas

    def escaped_name(self):
        name = pathlib.Path(self._name).name
        if sem.ON_WINDOWS:
            return re.sub(r'[:\\?"<>|]', "", name)
        else:
            return name

    def get_tokens(self):
        tokens = []
        content = self.content
        for span in self.segmentation("tokens"):
            tokens.append(content[span.lb: span.ub])
        return tokens

    def set_content(self, content):
        self._content = content

    def add_segmentation(self, segmentation):
        self._segmentations[segmentation.name] = segmentation
        self._segmentations[segmentation.name]._document = self

    def segmentation(self, name):
        return self._segmentations.get(name, None)

    def add_annotationset(self, annotationset):
        self._annotationsets[annotationset.name] = annotationset
        self._annotationsets[annotationset.name]._document = self

    def annotationset(self, name):
        return self._annotationsets.get(name, None)

    def add_metadata(self, key, value):
        self._metadatas[key] = value

    def metadata(self, name):
        return self._metadatas.get(name, None)

    def mime_type(self):
        return self.metadata("MIME")

    def write(self, f, depth=0, indent=4, add_header=False):
        if add_header:
            f.write('<?xml version="1.0" encoding="{0}" ?>\n'.format(f.encoding or "ASCII"))
        f.write(
            '{0}<document name="{1}">\n'.format(
                depth * indent * " ", self.name.replace('"', "&quot;")
            )
        )
        depth += 1
        f.write("{}<metadata".format(depth * indent * " "))
        for metakey, metavalue in sorted(self.metadatas.items()):
            f.write(' {0}="{1}"'.format(metakey, metavalue))
        f.write(" />\n")
        f.write(
            "{0}<content>{1}</content>\n".format(depth * indent * " ", html.escape(self.content))
        )

        if len(self.segmentations) > 0:
            f.write("{0}<segmentations>\n".format(depth * indent * " "))
            refs = [seg.reference for seg in self.segmentations.values() if seg.reference]
            for seg in sorted(
                self.segmentations.values(),
                key=lambda x: (x.reference is not None and x.reference.reference in refs, x.name),
            ):
                depth += 1
                ref = (
                    seg.reference.name if isinstance(seg.reference, Segmentation) else seg.reference
                )
                ref_str = "" if ref is None else ' reference="{0}"'.format(ref)
                f.write(
                    '{0}<segmentation name="{1}"{2}>'.format(
                        depth * indent * " ", seg.name, ref_str
                    )
                )
                depth += 1
                for i, element in enumerate(seg):
                    lf = i == 0 or (i % 5 == 0)
                    if lf:
                        f.write("\n{0}".format(depth * indent * " "))
                    f.write(
                        '{0}<s s="{1}" l="{2}" />'.format(
                            ("" if lf else " "), element.lb, len(element)
                        )
                    )
                f.write("\n")
                depth -= 1
                f.write("{0}</segmentation>\n".format(depth * indent * " "))
                depth -= 1
            f.write("{0}</segmentations>\n".format(depth * indent * " "))

        if len(self.annotationsets) > 0:
            f.write("{0}<annotations>\n".format(depth * indent * " "))
            for annotation in self.annotationsets.values():
                depth += 1
                reference = (
                    ""
                    if not annotation.reference
                    else ' reference="{0}"'.format(
                        annotation.reference
                        if isinstance(annotation.reference, str)
                        else annotation.reference.name
                    )
                )
                f.write(
                    '{0}<annotation name="{1}"{2}>\n'.format(
                        depth * indent * " ", annotation.name, reference
                    )
                )
                depth += 1
                for tag in annotation:
                    f.write(
                        '{0}<tag v="{1}" s="{2}" l="{3}"/>\n'.format(
                            depth * indent * " ", tag.getValue(), tag.lb, len(tag)
                        )
                    )
                depth -= 1
                f.write("{0}</annotation>\n".format(depth * indent * " "))
                depth -= 1
            f.write("{0}</annotations>\n".format(depth * indent * " "))

        depth -= 1
        f.write("{0}</document>\n".format(depth * indent * " "))

    def set_reference(
        self, annotation_name, reference_name, add_to_corpus=False, filter=get_top_level
    ):
        annot = self.annotationset(annotation_name)

        if annot is not None and (
            annot.reference is None or annot.reference.name != reference_name
        ):
            spans = self.segmentation(reference_name).get_reference_spans()
            begin = 0
            i = 0
            for annotation in annot:
                start = annotation.lb
                end = annotation.ub
                while not (spans[i].lb <= start and start < spans[i].ub):
                    i += 1
                begin = i
                while spans[i].ub < end:
                    i += 1
                annotation.lb = begin
                annotation.ub = i + 1
                i = max(begin - 1, 0)
                begin = 0
            annot._reference = self.segmentation(reference_name)

        if add_to_corpus:
            self.add_to_corpus(annotation_name, filter=filter)

    def add_to_corpus(self, annotation_name, filter=get_top_level, scheme="BIO"):
        base_annotations = self.annotationset(annotation_name) or AnnotationSet(annotation_name)
        annotations = base_annotations.get_reference_annotations()

        spans = self.segmentation("tokens").get_reference_spans()
        begin = 0
        i = 0
        to_remove = []  # annotations that cannot be aligned with tokens will be removed
        for j, annotation in enumerate(annotations):
            start = annotation.lb
            end = annotation.ub
            while (i > 0) and start < spans[i].lb:
                i -= 1
            while (i < len(spans)) and not (spans[i].lb <= start < spans[i].ub):
                i += 1
            if i < len(spans):
                begin = i
                while spans[i].ub < end:
                    i += 1
                annotation.lb = begin
                annotation.ub = i + 1
            else:
                mention = self.content[annotation.lb: annotation.ub].strip().replace("\n", " ")
                if len(mention) > 32:
                    mention = mention[:27] + "[...]"
                sem.logger.warning(
                    "document {0}, cannot add annotation {1}: {2}".format(
                        self.name, annotation, mention
                    )
                )
                to_remove.append(j)
            i = max(begin, 0)
            begin = 0

        for i in to_remove[::-1]:
            del annotations[i]

        if filter:
            annotations = filter(annotations)

        sentence_spans = iter(self.segmentation("sentences"))
        annot_index = 0
        shift = 0
        if len(annotations) == 0:
            annots = []
            cur_annot = None
        else:
            annots = annotations
            cur_annot = annots[annot_index]

        flags = chunking_schemes[scheme.upper()]
        BEGIN = flags["begin"]
        IN = flags["in"]
        LAST = flags["last"]
        SINGLE = flags["single"]
        OUT = flags["out"]
        for sentence in self.corpus.sentences:
            span = next(sentence_spans)
            tags = [OUT for _ in range(len(sentence))]
            while cur_annot is not None and cur_annot.lb >= span.lb and cur_annot.ub <= span.ub:
                if len(cur_annot) == 1:
                    tags[cur_annot.lb - shift] = "{}-{}".format(SINGLE, cur_annot.value)
                else:
                    tags[cur_annot.lb - shift] = "{}-{}".format(BEGIN, cur_annot.value)
                    for k in range(cur_annot.lb + 1, cur_annot.ub - 1):
                        tags[k - shift] = "{}-{}".format(IN, cur_annot.value)
                    tags[cur_annot.ub - 1 - shift] = "{}-{}".format(LAST, cur_annot.value)

                try:
                    annot_index += 1
                    cur_annot = annots[annot_index]
                except IndexError:
                    cur_annot = None

            sentence.add(tags, annotation_name)
            if cur_annot is not None and (cur_annot.lb in span and cur_annot.ub > span.ub):
                # annotation spans over at least two sentences
                mention = self.content[cur_annot.lb: cur_annot.ub].strip().replace("\n", " ")
                if len(mention) > 32:
                    mention = mention[:27] + "[...]"
                sem.logger.warning(
                    "document {}, annotation {} spans over multiple sentences, ignoring: {}".format(
                        self.name, annotation, mention
                    )
                )
                try:
                    annot_index += 1
                    cur_annot = annots[annot_index]
                except IndexError:
                    cur_annot = None
            shift += len(sentence)
        self.corpus.fields.append(annotation_name)

    def add_annotation_from_tags(self, tags, field, annotation_name):
        BIO = all([tag[0] in "BIO" for tag in tags[0]])
        if self._annotationsets.get(annotation_name, None):
            del self._annotationsets[annotation_name]._annotations[:]
        if BIO:
            self.add_chunking(tags, field, annotation_name)
        else:
            self.add_tagging(tags, field, annotation_name)
        if field not in self.corpus:
            self.corpus.fields += [field]

    def add_tagging(self, sentence_tags, field, annotation_name):
        nth_token = 0
        annotation = []

        for nth_sentence, tags in enumerate(sem.util.correct_pos_tags(sentence_tags)):
            index = len(annotation)
            i = len(tags) - 1
            n = 0
            current = None  # current tag value (for multiword tags)
            while i >= 0:
                if tags[i][0] != "_":
                    annotation.insert(index, Tag(tags[i], nth_token + i, 0, length=n + 1))
                    current = None
                    n = 0
                else:
                    if current is None:
                        current = tags[i].lstrip("_")
                        n = 0
                    n += 1
                i -= 1
            self.corpus.sentences[nth_sentence].add(tags, field)
            nth_token += len(tags)
        self._annotationsets[annotation_name] = AnnotationSet(
            annotation_name, reference=self.segmentation("tokens")
        )
        self._annotationsets[annotation_name].annotations = annotation[:]

    def add_chunking(self, sentence_tags, field, annotation_name):
        for nth_sentence, tags in enumerate(sentence_tags):
            self.corpus.sentences[nth_sentence].add(tags, field)
        self._annotationsets[annotation_name] = chunk_annotation_from_corpus(
            self.corpus, field, annotation_name, reference=self.segmentation("tokens")
        )


class SEMCorpus:
    def __init__(self, documents=None):
        if documents is None:
            self._documents = []
        else:
            self._documents = documents

    def __getitem__(self, index):
        return self._documents[index]

    def __len__(self):
        return len(self._documents)

    def __iter__(self):
        return iter(self._documents)

    @property
    def documents(self):
        return self._documents

    def add_document(self, document):
        ok = not any([d.name == document.name for d in self.documents])
        if ok:
            self._documents.append(document)

    def write(self, f, indent=4):
        f.write('<?xml version="1.0" encoding="{0}" ?>\n'.format(f.encoding or "ASCII"))
        f.write("<sem>\n")
        for document in self._documents:
            document.write(f, depth=1, indent=indent, add_header=False)
        f.write("</sem>")


str2docfilter = {
    "all documents": lambda x, y: True,
    "only documents with annotations": lambda d, a: len(d.annotationset(a) or []) > 0,
}


class Trie:
    """The Trie object.

    Attributes
    ----------
    _data : dict
        the structure where all the entries of a multiword dictionary
        are loaded.
    """

    def __init__(self, filename=None, encoding=None):
        self._data = {}

        if filename:
            encoding = encoding or "UTF-8"
            for line in open(filename, "rU", encoding=encoding):
                seq = line.strip().split()

                self.add(seq)

    def __iter__(self):
        seq = []

        def dfs(dic):
            """Depth First Search"""
            keys = set(dic.keys())
            found = NUL in keys

            if found:
                keys.remove(NUL)
                if dic[NUL]:
                    yield seq
            keys = list(keys)
            keys.sort()
            for k in keys:
                seq.append(k)
                for i in dfs(dic[k]):
                    yield i
                seq.pop()

        for i in dfs(self._data):
            yield i

    def __len__(self):
        length = 0
        for _ in self:
            length += 1
        return length

    @property
    def data(self):
        return self._data

    def add(self, sequence):
        iterator = sequence.__iter__()
        d = self._data

        try:
            while True:
                token = next(iterator)

                if token not in d:
                    d[token] = {}

                d = d[token]
        except StopIteration:
            pass

        d[NUL] = {}

    def add_with_value(self, sequence, value):
        iterator = iter(sequence)
        d = self._data

        try:
            while True:
                token = next(iterator)

                if token not in d:
                    d[token] = {}

                d = d[token]
        except StopIteration:
            pass

        d[NUL] = value

    def contains(self, sequence):
        iterator = iter(sequence)
        d = self._data
        result = True

        try:
            while True:
                token = next(iterator)

                if token not in d:
                    result = False
                    break

                d = d[token]
        except StopIteration:
            pass

        return result and (NUL in d)

    def remove(self, sequence):
        def remove(dic, iterator):
            try:
                elt = next(iterator)
                if elt in dic:
                    remove(dic[elt], iterator)
                    if dic[elt] == {}:
                        del dic[elt]
            except StopIteration:
                if NUL in dic:
                    del dic[NUL]

        remove(self._data, iter(sequence))

    def goto(self, sequence):
        iterator = iter(sequence)
        d = self._data
        result = True

        try:
            while True:
                token = next(iterator)

                if token not in d:
                    result = False
                    break

                d = d[token]
        except StopIteration:
            pass

        if result:
            return d
        else:
            return None


def Coder():
    import warnings
    warnings.warn("sem.storage.Coder is deprecated, use sem.CRF.Quark instead", DeprecationWarning)
    return Quark()



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
file: __main__.py

Description: the entry point to SEM.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import pathlib
import unittest
import sys

import sem
import sem.modules
from sem.util import find_suggestions
from sem.tests import (test_features, test_processors, test_segmentation, test_workflows)


def main(argv=None):
    def banter():
        def username():
            import os

            return os.environ.get("USERNAME", os.environ.get("USER", pathlib.Path.home().name))

        import random

        banters = [
            "Do thou mockest me?",
            "Try again?",
            "I'm sorry {0}, I'm afraid I can't do that.".format(username()),
            "The greatest trick this module ever pulled was convincing the users it did not exist.",
            "It's just a typo.",
        ]
        return random.choice(banters)

    modules = {}
    for modulename in sem.modules.names():
        module = sem.modules.get_module(modulename)
        if hasattr(module, "main"):
            modules[modulename] = module
    name = pathlib.Path(sys.argv[0]).name
    operation = sys.argv[1] if len(sys.argv) > 1 else "-h"

    if operation in modules:
        module = modules[operation]
        module.main(sys.argv[2:])
    elif operation in ["-h", "--help"]:
        print("Usage: {0} <module> [module arguments]\n".format(name))
        print("Module list:")
        for module in sorted(modules):
            print("\t{0}".format(module))
        print()
        print("for SEM's current version: -v or --version\n")
        print("for informations about SEM: see readme.md")
        print("for playing all tests: --test")
    elif operation in ["-v", "--version"]:
        print(sem.full_name())
    elif operation == "--test":
        loader = unittest.TestLoader()
        testsuite = loader.loadTestsFromModule(test_features)
        testsuite.addTests(loader.loadTestsFromModule(test_processors))
        testsuite.addTests(loader.loadTestsFromModule(test_segmentation))
        testsuite.addTests(loader.loadTestsFromModule(test_workflows))
        unittest.TextTestRunner(verbosity=2).run(testsuite)
    else:
        print("Module not found: {}".format(operation))
        suggestions = find_suggestions(operation, modules)
        if len(suggestions) > 0:
            print("Did you mean one of the following?")
            for suggestion in suggestions:
                print("\t{0}".format(suggestion))
        else:
            print("No suggestions found...", banter())


if __name__ == "__main__":
    main()



# -*- coding:utf-8 -*-

"""
file: export.py

Description: exports an input to a given format.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse

# measuring time laps
import time
from datetime import timedelta

import sem.importers
from sem.processors import ExportProcessor
from sem.util import str2bool

from sem.storage import SEMCorpus
import sem.logger


def main(argv=None):
    export(**vars(parser.parse_args(argv)))


def export(
    infile,
    exporter_name,
    outfile,
    pos_column=None,
    chunk_column=None,
    ner_column=None,
    lang="fr",
    lang_style="default.css",
    import_options=None,
    ienc=None,
    oenc=None,
    enc="utf-8",
):
    start = time.time()

    ienc = ienc or enc
    oenc = oenc or enc

    exporter = ExportProcessor(
        exporter_name,
        lang=lang,
        lang_style=lang_style,
        pos_column=pos_column,
        chunk_column=chunk_column,
        ner_column=ner_column,
    )

    if type(import_options) in (list,):  # list from argparse
        options = {}
        for option in import_options:
            key, value = option.split("=", 1)
            try:
                value = str2bool(value)
            except ValueError:
                pass
            if ',' in value:
                try:
                    value = [item for item in value.split(',') if item]
                except AttributeError:
                    pass
            options[key] = value
        options["encoding"] = ienc
    else:
        options = import_options

    infile_is_str = isinstance(infile, str)
    if infile_is_str:
        sem.logger.info("loading input file")
        document = sem.importers.load(infile, **options)
        if isinstance(document, SEMCorpus):
            sem.logger.warning("input file is SEM corpus, only exporting the first document")
            document = document[0]
    else:
        sem.logger.info("using input document")
        document = infile

    sem.logger.debug("exporting document {0}".format(document.name))
    exporter.process_document(document, outfile, encoding=oenc)

    laps = time.time() - start
    sem.logger.info("done in %s", timedelta(seconds=laps))


parser = argparse.ArgumentParser("Export CoNLL-formatted data to specified format.")

parser.add_argument("infile", help="The input file")
parser.add_argument("exporter_name", help="The name of the exporter to use")
parser.add_argument("outfile", help="The output file")
parser.add_argument(
    "-p", "--pos-column", dest="pos_column", default=None, help="The column for POS."
)
parser.add_argument(
    "-c", "--chunk-column", dest="chunk_column", default=None, help="The column for chunk."
)
parser.add_argument(
    "-n", "--ner-column", dest="ner_column", default=None, help="The column for NER."
)
parser.add_argument(
    "--lang", dest="lang", default="fr", help="The language of the text (default: %(default)s)"
)
parser.add_argument(
    "-s",
    "--lang-style",
    dest="lang_style",
    default="default.css",
    help="The style to use, if applicable (default: %(default)s)",
)
parser.add_argument(
    "--import-options",
    nargs="*",
    dest="import_options",
    help=(
        'Import-specific options. This is a list of "key=value" pairs.'
        'Typical keys for this options are: "fields" (the list of CoNLL fields),'
        '"word_field" (the name of the field where words are), "taggings" (the fields'
        'that contains taggings to load as annotations), "chunkings" (IOB formatted'
        'tags that represent some chunking (syntax or NE).'
    ),
)
parser.add_argument("--input-encoding", dest="ienc", help="Encoding of the input (default: UTF-8)")
parser.add_argument("--output-encoding", dest="oenc", help="Encoding of the input (default: UTF-8)")
parser.add_argument(
    "--encoding",
    dest="enc",
    default="UTF-8",
    help="Encoding of both the input and the output (default: UTF-8)",
)
parser.add_argument(
    "-l",
    "--log",
    dest="log_level",
    choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
    default="WARNING",
    help="Increase log level (default: %(default)s)",
)
parser.add_argument("--log-file", dest="log_file", help="The name of the log file")



"""
file: tokenisers.py

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re

import sem.constants
from sem.storage import Span, add_last


spaces = re.compile(r"\s+", re.U + re.M)


def bounds2spans(bounds):
    """Create spans from bounds.
    """
    spans = [Span(bounds[i].ub, bounds[i + 1].lb) for i in range(0, len(bounds) - 1)]
    spans = [span for span in spans if span.lb != span.ub]
    return spans


class Tokeniser:
    """Root tokeniser class"""
    def __init__(self):
        # adding some "basic" things to not tokenise across languages.
        self._forbidden = [sem.constants.url_re, sem.constants.email_re]
        self._force = []

    def word_spans(self, content):
        l1 = [match.span() for match in spaces.finditer(content)]
        l2 = [(l1[i][1], l1[i + 1][0]) for i in range(len(l1) - 1)]

        if l1[0][0] != 0:
            l2.insert(0, (0, l1[0][0]))
        if l1[-1][1] != len(content):
            l2.append((l1[-1][1], len(content)))

        return [Span(span[0], span[1]) for span in l2]

    def sentence_bounds(self, content, token_spans):
        """Return a list of bounds matching sentences.

        Parameters
        ----------
        content : str
            the content to get sentence bounds for.
        token_spans : list of Span
            the list of tokens spans.

        Returns
        -------
        list[span]
            The spans of the sentence bounds.
        """

        sent_bounds = [Span(0, 0)]
        for index, span in enumerate(token_spans):
            token = content[span.lb: span.ub]
            if token in "\r\n":
                add_last(sent_bounds, Span(index, index + 1))
        add_last(sent_bounds, Span(len(token_spans), len(token_spans)))

        return sent_bounds

    def sentence_spans(self, content, token_spans):
        return bounds2spans(self.sentence_bounds(content, token_spans))

    def paragraph_bounds(self, content, sentence_spans, token_spans):
        """Return a list of bounds matching paragraphs.

        Parameters
        ----------
        content : str
            the content to find paragraph bounds for.
        sentence_spans : list[Span]
            the list of sentence spans.
        token_spans : list[Span]
            the list of token spans.

        Returns
        -------
        list[Span]
            The list of paragraph bounds in content.
        """

        s_spans = [Span(token_spans[e.lb].lb, token_spans[e.ub - 1].ub) for e in sentence_spans]
        bounds = [Span(0, 0)]
        for index, sentence in enumerate(sentence_spans[1:], 1):
            substring = content[s_spans[index - 1].ub: s_spans[index].lb]
            if substring.count("\n") > 1:
                bounds.append(Span(index, index))
        bounds.append(Span(len(sentence_spans), len(sentence_spans)))

        return bounds

    def paragraph_spans(self, content, sentence_spans, token_spans):
        return bounds2spans(self.paragraph_bounds(content, sentence_spans, token_spans))


class FrenchTokeniser(Tokeniser):
    def __init__(self):
        self._dots = re.compile(r"(\.{2,})$")
        self._cls = re.compile(r"(-je|-tu|-nous|-vous|(:?-t)?-(:?on|ils?|elles?))\b", re.U + re.I)
        self._is_abn = re.compile(r"\b(dr|me?lles?|mme?s?|mr?s?|st)\.?$", re.U + re.I)
        self._abbrev = re.compile(r"\b(i\.e\.|e\.g\.|c-à-d)", re.U + re.I)
        self._digit_valid = set("0123456789,.-")
        self._simple_smileys = re.compile("^[:;x=],?-?[()DdPp]+$")

        self._forbidden = [self._is_abn, self._abbrev, self._simple_smileys]

        self._force = [sem.constants.url_re, sem.constants.email_re, self._cls]

        self._word = re.compile("^[^\\W\\d]+$", re.U + re.M)
        self._number_with_unit = re.compile("([0-9][^0-9,.])|([^0-9,.][0-9])")
        self._atomic = re.compile('[;:«»()\\[\\]{}=+*$£€/\\"?!…%€$£]')
        self._comma_not_number = re.compile("(?<=[^0-9]),(?![0-9])", re.U + re.M)
        self._apostrophe = re.compile("(?=['ʼ’])", re.U + re.M)

    def word_spans(self, content):
        spans = []
        offset = 0

        part = content
        l1 = [match.span() for match in spaces.finditer(part)]

        if l1:
            l2 = [(l1[i][1], l1[i + 1][0]) for i in range(len(l1) - 1)]
            if l1[0][0] != 0:
                l2.insert(0, (0, l1[0][0]))
            if l1[-1][1] != len(part):
                l2.append((l1[-1][1], len(part)))
        else:
            l2 = [(0, len(part))]

        i = 0
        while i < len(l2):
            span = l2[i]
            text = part[span[0]: span[1]]
            shift = span[0]
            if len(text) == 1:
                i += 1
                continue
            if self._word.match(text):
                i += 1
                continue
            found = False
            for force in self._force:
                found = force.search(text)
                if found:
                    s, e = found.start(), found.end()
                    del l2[i]
                    l2.insert(i, (shift + e, shift + len(text)))
                    l2.insert(i, (shift + s, shift + e))
                    l2.insert(i, (shift, shift + s))
                    i += 2
                    break
            if found:
                continue
            for forbidden in self._forbidden:
                found = forbidden.match(text)
                if found:
                    i += 1
                    break
            if found:
                continue
            tmp = []
            # atomic characters, they are always split
            prev = span[0]
            for find in self._atomic.finditer(text):
                if prev != shift + find.start():
                    tmp.append((prev, shift + find.start()))
                tmp.append((shift + find.start(), shift + find.end()))
                prev = shift + find.end()
            if tmp:
                if prev != span[1]:
                    tmp.append((prev, span[1]))
                del l2[i]
                for t in reversed(tmp):
                    l2.insert(i, t)
                continue
            # commas
            prev = span[0]
            find = self._comma_not_number.search(text)
            if find:
                tmp.extend(
                    [
                        (prev, shift + find.start()),
                        (shift + find.start(), shift + find.end()),
                        (shift + find.end(), span[1]),
                    ]
                )
                prev = shift + find.end() + 1
            if tmp:
                del l2[i]
                for t in reversed(tmp):
                    l2.insert(i, t)
                continue
            # apostrophes
            prev = span[0]
            for find in self._apostrophe.finditer(text):
                tmp.append((prev, shift + find.start() + 1))
                prev = shift + find.start() + 1
            if prev < span[1]:
                tmp.append((prev, span[1]))
            if len(tmp) > 1:
                del l2[i]
                for t in reversed(tmp):
                    l2.insert(i, t)
                continue
            del tmp[:]
            # number with unit
            prev = span[0]
            for find in self._number_with_unit.finditer(text):
                tmp.append((prev, span[0] + find.start() + 1))
                prev = span[0] + find.start() + 1
            if tmp:
                tmp.append((prev, span[1]))
                del l2[i]
                for t in reversed(tmp):
                    l2.insert(i, t)
                continue
            # dots and ending commas
            if text and (text[-1] in ".," and not (len(text) == 2 and text[0].isupper())):
                mdots = self._dots.search(text)
                length = len(mdots.group(1)) if mdots else 1
                if length != len(text):
                    tmp = [(span[0], span[1] - length), (span[1] - length, span[1])]
            if tmp:
                del l2[i]
                for t in reversed(tmp):
                    l2.insert(i, t)
                continue
            i += 1

        spans = [Span(s[0] + offset, s[1] + offset) for s in l2 if s[0] < s[1]]
        spans = [span for span in spans if len(span) > 0]
        return spans

    def sentence_bounds(self, content, token_spans):
        sent_bounds = []
        tokens = [content[t.lb: t.ub] for t in token_spans]
        opening_counts = [0 for i in token_spans]
        count = 0
        for i in range(len(opening_counts)):
            if tokens[i] in "«([":
                count += 1
            elif tokens[i] in "»)]":
                count -= 1
            opening_counts[i] = count

        sent_bounds.append(Span(0, 0))
        for index, span in enumerate(token_spans):
            token = tokens[index]
            if re.match("^[?!]+$", token) or token == "…" or re.match("\\.\\.+", token):
                sent_bounds.append(Span(index + 1, index + 1))
            elif token == ".":
                if opening_counts[index] == 0:
                    sent_bounds.append(Span(index + 1, index + 1))
            elif (
                index < len(token_spans) - 1
                and content[span.ub: token_spans[index + 1].lb].count("\n") > 1
            ):
                sent_bounds.append(Span(index + 1, index + 1))
        sent_bounds.append(Span(len(tokens), len(tokens)))

        return sent_bounds


class EnglishTokeniser(Tokeniser):
    def word_spans(self, s):
        bounds = [Span(0, 0)]

        atomic = set(';:«»()[]{}=+*$£€/\\"?!%€$£')
        apostrophe = set("'ʼ’")

        for index, c in enumerate(s):
            is_first = index == 0
            is_last = index == len(s) - 1
            if c.isspace():
                add_last(bounds, Span(index, index + 1))
            elif c in atomic:
                add_last(bounds, Span(index, index))
                bounds.append(Span(index + 1, index + 1))
            elif c in apostrophe:
                if is_first or is_last:
                    add_last(bounds, Span(index, index))
                    bounds.append(Span(index + 1, index + 1))
                elif s[index + 1] == s[index]:
                    bounds.append(Span(index, index + 1))
                else:
                    if s[index - 1] == "n" and s[index + 1] == "t":
                        bounds.append(Span(index - 1, index - 1))
                        bounds.append(Span(index + 2, index + 2))
                    elif s[index + 1] == "s":
                        bounds.append(Span(index, index))
                        bounds.append(Span(index + 2, index + 2))
                    else:
                        add_last(bounds, Span(index, index))
            elif c in ".,":
                if is_first or is_last:
                    add_last(bounds, Span(index, index))
                    bounds.append(Span(index + 1, index + 1))
                elif (is_first or not s[index - 1].isdigit()) and (
                    is_last or not s[index - 1].isdigit()
                ):
                    add_last(bounds, Span(index, index))
                    bounds.append(Span(index + 1, index + 1))

        bounds.append(Span(len(s), len(s)))

        return bounds2spans(bounds)

    def sentence_bounds(self, content, token_spans):
        sent_bounds = []
        tokens = [content[t.lb: t.ub] for t in token_spans]
        openings = set(["«", "(", "[", "``"])
        closings = set(["»", ")", "]", "''"])
        opening_counts = [0 for i in tokens]
        count = 0
        for i in range(len(opening_counts)):
            if tokens[i] in openings:
                count += 1
            elif tokens[i] in closings:
                count -= 1
            opening_counts[i] = count

        sent_bounds.append(Span(0, 0))
        for index, token in enumerate(tokens):
            if re.match("^[?!]+$", token) or token == "…" or re.match(r"\.\.+", token):
                sent_bounds.append(Span(index + 1, index + 1))
            elif token == ".":
                if opening_counts[index] == 0:
                    sent_bounds.append(Span(index + 1, index + 1))
        sent_bounds.append(Span(len(tokens), len(tokens)))

        return sent_bounds


__tokenisers = {
    "default": Tokeniser,
    "fr": FrenchTokeniser,
    "en": EnglishTokeniser
}


def get_tokeniser(name):
    """Return the right tokeniser given name.

    Parameters
    ----------
    name : str
        the name of the tokeniser

    Returns
    -------
    Tokeniser
        The tokeniser matching name
    """
    return __tokenisers[name]
